#
# Functions to process '.nsb' files
# File created by Dave Dykstra <dwd@bell-labs.com>, 22 Nov 1996
#
#
# Copyright (C) 1997 by Dave Dykstra and Lucent Technologies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#  
# If those terms are not sufficient for you, contact the author to 
# discuss the possibility of an alternate license.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

#
# make a '.nsb' contents table, starting from the package description
#   contents table
#

proc makeNsbContents {npdContentsName nsbContentsName} {
    upvar $npdContentsName npdContents
    upvar $nsbContentsName nsbContents
    catch {unset nsbContents}

    validatePackageName $npdContents(package)

    # Note to dwd: do pathSubs and regSubs need checking here, or
    #  should they just be left to be handled on the other end?

    global npdNsbKeys cfgContents
    applyKeys $npdNsbKeys npdContents nsbContents nsb

    global knownMdTypes
    if {[info exists cfgContents(mdType)]} {
	set mdType $cfgContents(mdType)
	if {[lsearch -exact $knownMdTypes $mdType] < 0} {
	    nsbderror "unknown mdType: $mdType; should be one of $knownMdTypes"
	}
    } else {
	set mdType [lindex $knownMdTypes 0]
    }

    set nsbContents(generatedBy) [nsbdVersion]
    set nsbContents(generatedAt) [currentTime]


    if {[info exists npdContents(localTop)]} {
	set localTop $npdContents(localTop)
    } else {
	set localTop "."
    }
    if {![file isdirectory $localTop]} {
	nsbderror "localTop is not an existing directory: $localTop"
    }
    set globTop [glob $localTop]
    if {[llength $globTop] > 1} {
	nsbderror "localTop may not contain wildcards or spaces: $localTop"
    }
    # this will canonicalize the directory name, with one slash at the end
    set localTop $globTop/
    set relStart [string length $localTop]

    if {![info exists npdContents(paths)]} {
	warnmsg "Warning: no 'paths' keyword found"
	return
    }
    if {![info exists npdContents(nsbUrl)]} {
	warnmsg "Warning: no 'nsbUrl' keyword found"
    }
    if {![info exists npdContents(topUrl)] &&
		![info exists npdContents(multigetUrl)]} {
	nsbderror "no topUrl or multigetUrl keyword found"
    }
    if {[info exists npdContents(excludePaths)]} {
	set excludePaths $npdContents(excludePaths)
    } else {
	set excludePaths ""
    }
    set numerrors 0
    foreach path $npdContents(paths) {
	if {[set msg [relativePathCheck $path]] != ""} {
	    nsbpatherror $msg
	    continue
	}

	#
	# Expand %V
	#
	if {[regexp %V $path]} {
	    if {[info exists npdContents(version)]} {
		regsub -all %V $path "$npdContents(version)" path
	    } else {
		regsub -all %V $path "" path
	    }
	}
	#
	# Expand %E to glob expression
	#
	if {[regexp %E $path]} {
	    if {[info exists npdContents(executableTypes)]} {
		regsub -all %E $path "{[join $npdContents(executableTypes) ,]}" path
	    } else {
		regsub -all %E $path "" path
	    }
	}

	set paths [glob -nocomplain "$localTop$path"]
	if {$paths == ""} {
	    nsbpatherror "path not found: $localTop$path"
	    continue
	}
	foreach exppath $paths {
	    nsbexpandpath $mdType $exppath $relStart $excludePaths nsbContents
	}
    }
    if {$numerrors > 0} {
	if {$numerrors == 1} {
	    nsbderror "1 path error"
	} else {
	    nsbderror "$numerrors path errors"
	}
    }
}

#
# keep a count of the number of errors, print message, and continue
#   so all errors can be located in one pass
#
proc nsbpatherror {str} {
    upvar numerrors numerrors
    warnmsg $str
    incr numerrors
}

#
# expand the path into the nsbContents.  relStart is the string index
#  of the relative portion of the path.
#
proc nsbexpandpath {mdType path relStart excludePaths nsbContentsName} {
    upvar numerrors numerrors
    upvar pathTable pathTable
    upvar $nsbContentsName nsbContents
    set relPath [string range $path $relStart end]
    foreach excludePath $excludePaths {
	if {[string match $excludePath $relPath]} {
	    return
	}
    }
    if {[catch {file readlink $path} link] == 0} {
	# symbolic link.  If the link is to another path that *could be* in
	#   the package (i.e., under the same relative base) and it is a
	#   relative path itself, include it as a linkTo.  Probably the user
	#   would prefer to only have it remembered as a linkTo if the path
	#   *is* in the package, but that's much harder to implement.
	if {[string index $link 0] != "/"} {
	    set link [cleanPath [file join [file dirname $relPath] $link]]
	    # if remaining link does not begin with ../, it is below
	    #   the base so remember it as a 'linkTo'
	    if {[string range $link 0 2] != "../"} {
		if {![nsbaddpath $relPath nsbContents]} return
		lappend nsbContents([list paths $relPath linkTo]) $link
		return
	    }
	}
    }
    if {[file isdirectory $path]} {
	debugmsg "Including all files in directory $relPath"
	if {![nsbaddpath $relPath/ nsbContents]} return
	foreach exppath [glob $path/{.*,*}] {
	    set tail [file tail $exppath]
	    if {($tail != ".") && ($tail != "..")} {
		nsbexpandpath $mdType $exppath $relStart $excludePaths nsbContents
	    }
	}
    } else {
	set code [catch {set fd [open $path "r"]} string]
	if {$code != 0} {
	    nsbpatherror "$string"
	    return
	}
	fconfigure $fd -translation binary
	set code [catch {set mdList [$mdType -chan $fd]} string]
	close $fd
	if {$code != 0} {
	    nsbpatherror "error during $mdType of $path: $string"
	    return
	}

	if {![nsbaddpath $relPath nsbContents]} return
	set nsbContents([list paths $relPath length]) [lindex $mdList 0]
	set nsbContents([list paths $relPath $mdType]) [lindex $mdList 1]
	set mode "r"
	if {[file writable $path]} {
	    set mode "${mode}w"
	}
	if {[file executable $path]} {
	    set mode "${mode}x"
	}
	if {$mode != "rw"} {
	    set nsbContents([list paths $relPath mode]) $mode
	}
    }
}

#
# add a path to nsbContents(paths), if it was not already there
# return 1 if successful, otherwise 0
#
proc nsbaddpath {path nsbContentsName} {
    upvar $nsbContentsName nsbContents
    upvar pathTable pathTable
    upvar numerrors numerrors
    if {[info exists pathTable($path)]} {
	nsbpatherror "duplicate path $path"
	return 0
    }
    set pathTable($path) ""
    lappend nsbContents(paths) $path
    return 1
}

#
# check to make sure path is a relative path.
# If it is, return an empty string, otherwise return an error message
#
proc relativePathCheck {path} {
    set firstc [string index $path 0]
    if {$firstc == "/"} {
	return "path may not begin with slash: $path"
    }
    if {$firstc == "~"} {
	return "path may not begin with tilde: $path"
    }
    if {(($firstc == ".") && ([string index $path 1] == ".")) ||
		[regexp {/\.\./} "/$path/"]} {
	return ".. components not allowed in path: $path"
    }
    if {[regexp {{.*}} $path]} {
	# curly brackets would have an effect on glob which could defeat
	#  above checks; don't want to support them anyway because it
	#  would make reimplementation in a language other than tcl more
	#  difficult
	return "curly brackets not allowed in path: $path"
    }
    return ""
}

#
# Clean up the path: remove multiple slashes, and as many "." and ".."
#   components in the path as possible, without changing the file referenced. 
# Note: paths starting with /.. do not get collapsed to / like they would on
#   unix, and a trailing slash does not get removed.
#
proc cleanPath {path} {
    while {[regsub "//" $path "/" path]} {}
    set path "/$path/"
    regsub -all {/\./} $path "/" path
    while {[regsub {/[^/][^/]*/\.\./} $path "/" path]} {}
    set path [string range $path 1 [expr [string length $path] - 2]]
    if {$path == ""} {
	return "."
    }
    return $path
}

#
# validate a package name value
#

proc validatePackageName {name} {
    # '+' may be another character desired in a package name, but
    #   I want to reserve it in case someday I want to have the ability
    #   to add to keyword lists on the command line as opposed to
    #   overriding the lists with '='.
    # If you change the list here, also change the error message here and
    #   change the comments on the 'package' keys in tables.tcl.
    if {![regexp {^[A-Za-z0-9!_-]+$} $name]} {
	nsbderror "package name must contain only alphanumeric or \[!_-\]: $name"
    }
}

#
# Process a '.nsb' Not-So-Bad file
#
proc procnsbfile {fd lineno filename expectedPackageName hassig nsbScratch} {
    global procNsbType

    set package $expectedPackageName
    set sigtime ""

    global nsbContents
    catch {unset nsbContents}
    if {$procNsbType == "fetchAll"} {
	# process any contents
	nsbdParseContents $fd nsb $lineno nsbContents
	if {![info exists nsbContents(package)]} {
	    nsbderror "no package keyword in file"
	}
	# fall through

    } elseif {$package == ""} {
	# process any contents, don't yet know the package name
	nsbdParseContents $fd nsb $lineno nsbContents

	if {![info exists nsbContents(package)]} {
	    nsbderror "no package keyword in file"
	}

    } else {
	# only this package is allowed.
	if {(($procNsbType == "poll") || ($procNsbType == "update") ||
			($procNsbType == "preview")) &&
	      ([set aStoredNsbfile [findStoredNsbfile $package]] != "")} {
	    # See if it is exactly the same as an old one
	    set catchcode [catch {compareFiles $nsbScratch \
					    $aStoredNsbfile} answer]
	    if {($catchcode == 0) && $answer} {
		updatemsg "No change in .nsb file for package $package"
		return
	    }
	    if {$catchcode > 0} {
		debugmsg "could not compare to saved .nsb because $answer"
	    }
	}

	global ignoreSecurity
	if {!$ignoreSecurity} {
	    # Make sure it has a valid signature for this package before trying
	    #  to parse it
	    if {!$hassig} {
		nsbderror "no PGP signature found; -ignoreSecurity required to process"
	    }
	    set maintainers [nrdLookup $package maintainers]
	    if {$maintainers == ""} {
		nsbderror "no authorized maintainers registered for package $package;\n  -ignoreSecurity required to process"
	    }
	    # check the PGP signature
	    set ans [pgpCheckFile $nsbScratch $filename]
	    foreach {goodsig user sigtime warning} $ans {}
	    if {$goodsig} {
		if {[lsearch -exact $maintainers $user] < 0} {
		    nsbderror "user ID \"$user\" not in authorized maintainers for package $package"
		}
		if {$warning != ""} {
		    warnmsg "Warning from pgp for $filename:\nWARNING: $warning\n"
		} else {
		    updatemsg "Good signature from $user"
		}
	    } else {
		# !goodsig
		if {$warning != ""} {
		    set warning "\n$warning"
		}
		if {$user == "*UNKNOWN*"} {
		    nsbderror "unrecognized key error from pgp$warning"
		}
		if {$user != ""} {
		    nsbderror "bad signature from pgp user ID \"$user\"$warning"
		}
		nsbderror "error from pgp$warning"
	    }
	} elseif {$hassig} {
	    updatemsg "Ignoring signature because of -ignoreSecurity"
	}

	# any signature at the end of the file will be
	#   ignored by nsbdParseContents
	nsbdParseContents $fd nsb $lineno nsbContents

	if {![info exists nsbContents(package)]} {
	    nsbderror "no package keyword in file"
	}

	if {$nsbContents(package) != $package} {
	    nsbderror "incorrect package $nsbContents(package), expected $package"
	}

	if {[info exists nsbContents(releaseNote)]} {
	    noticemsg "$package releaseNote:\n  [join $nsbContents(releaseNote) "\n  "]"
	}
    }

    #
    # Apply command line keyword settings, saving the executableTypes
    #   and version for later
    #
    if {[info exists nsbContents(executableTypes)]} {
	set savedExecutableTypes $nsbContents(executableTypes)
    }
    if {[info exists nsbContents(version)]} {
	set savedVersion $nsbContents(version)
    }
    applyCmdkeys nsbContents nsb


    #
    # Put out informational progress message
    #
    set package $nsbContents(package)
    validatePackageName $package
    if {![info exists nsbContents(generatedBy)]} {
	nsbderror "missing generatedBy field"
    }
    if {![info exists nsbContents(generatedAt)]} {
	nsbderror "missing generatedAt field"
    }
    updatemsg "$package '.nsb' file generated at $nsbContents(generatedAt)"
    updatemsg "  by $nsbContents(generatedBy)"


    #
    # Determine the applicable executableTypes and versions
    # There is only one version allowed in a single package description,
    #   but I turn it into a list for these reasons: 1) I had originally 
    #   supported multiple versions in a single package, 2) I may want
    #   to change it back someday, and 3) there are some cases when
    #   NSBD does support dealing with multiple versions of the same
    #   package, with multiple '.nsb' files, and some routines are shared.
    #
    set executableTypes [getExecutableTypes $package nsbContents]
    if {$executableTypes == ""} {
	set executableTypes [list ""]
    }
    if {[info exists nsbContents(version)]} {
	set versions [list $nsbContents(version)]
    } else {
	set versions [list ""]
    }
    #
    # Restore pre-command line keyword settings of executableTypes and
    #   version, if any, so will still be able to tell which were
    #   supported in the '.nsb' file.
    #
    if {[info exists savedExecutableTypes]} {
	set nsbContents(executableTypes) $savedExecutableTypes
    }
    if {[info exists savedVersion]} {
	set nsbContents(version) $savedVersion
    }

    if {$procNsbType == "fetchAll"} {
	progressmsg "Fetching all files for verification"
	fetchAllNsbContents $executableTypes $versions $nsbScratch
	progressmsg "All URLs in $filename OK"
	return
    }

    checkOSVersions $executableTypes

    if {$procNsbType == "register"} {
	if {[promptUnmetRequirements $package $executableTypes]} {
	    promptToRegister $package $executableTypes $versions
	}
	return
    }

    #
    # Continue in another routine, to avoid strings that are
    #   too long for tcl2c
    #
    procnsbfile2 $package $executableTypes $versions $sigtime $nsbScratch
}
    
proc procnsbfile2 {package executableTypes versions sigtime nsbScratch} {
    global procNsbType

    #
    # check required packages
    #
    set unmet [unmetRequirementPackages $executableTypes]
    if {$unmet != ""} {
	set msglist ""
	foreach {pack msg nsbUrls} $unmet {
	    lappend msglist "Warning: required package $pack $msg"
	}
	warnmsg [join $msglist "\n"]
    }

    #
    # If the package will be stored in multiple places because of multiple
    #   executableTypes, process separately for each place.   Note that if
    #   an executableType is removed from the package, files that had
    #   been installed for that old executableType will be left alone
    #

    #
    # Determine the unique installTops, and which executableTypes
    #  and nsbStoreNames apply to each.
    #
    set installTops ""
    set allNsbStoreNames ""
    foreach executableType $executableTypes {
	set executableType [cleanSubst $executableType]
	foreach version $versions {
	    set version [cleanSubst $version]
	    set installTop [getInstallTop $package $executableType $version]
	    if {![info exists topExecutableTypes($installTop)]} {
		lappend installTops $installTop
	    }
	    luniqueAppend topExecutableTypes($installTop) $executableType
	    luniqueAppend topVersions($installTop) $version
	    set nsbStoreName [getNsbStoreName $package $executableType $version]
	    # note that it doesn't make much sense to have multiple
	    #  nsbStoreNames for the same installTop, but it is possible
	    #  if a user specifies a %E or %V in the nsbStorePath but
	    #  doesn't specify a per-executableType or per-version installTop
	    if {![info exists topNsbStoreNames($installTop)] ||
		    [lsearch -exact $topNsbStoreNames($installTop) \
							$nsbStoreName] < 0} {
		lappend topNsbStoreNames($installTop) $nsbStoreName
		if {[lsearch -exact $allNsbStoreNames $nsbStoreName] >= 0} {
		    nsbderror \
    "May not store '.nsb' file in the same place for different installTops"
		}
		lappend allNsbStoreNames $nsbStoreName
	    }
	}
    }

    #
    # Process the '.nsb' file separately for each installTop
    #
    set anotherInstallTop ""
    global nupContents
    foreach top $installTops {
	if {$topExecutableTypes($top) == [list ""]} {
	    set topExecutableTypes($top) ""
	}
	if {$topVersions($top) == [list ""]} {
	    set topVersions($top) ""
	}
	if {$sigtime != ""} {
	    # compare the signature times on the old stored files
	    foreach nsbStoreName $topNsbStoreNames($top) {
		if {[file exists $nsbStoreName]} {
		    set ans [pgpCheckFile $nsbStoreName]
		    set oldtime [lindex $ans 2]
		    if {($oldtime != "") && \
				([compareTimes $oldtime $sigtime] > 0)} {
			nsbderror "PGP signature time ($sigtime) older than time on $nsbStoreName ($oldtime)"
		    }
		}
	    }
	}

	set hadPaths [makeNupContents [lindex $topNsbStoreNames($top) 0] \
		$nsbScratch $top $topExecutableTypes($top) $topVersions($top)]

	if {$procNsbType == "preview"} {
	    previewNupPaths
	    continue
	}

	fetchNupPaths $anotherInstallTop

	executeUpdateCommands $package "preUpdateCommands" nupContents

	set warndups 0
	foreach nsbStoreName $topNsbStoreNames($top) {
	    if {$hadPaths} {
		updatemsg "Saving $nsbStoreName"
		nsbdcatch {file mkdir [file dirname $nsbStoreName]}
		nsbdcatch {file copy -force $nsbScratch $nsbStoreName}
	    } elseif {[file exists $nsbStoreName]} {
		updatemsg "Deleting $nsbStoreName because no paths in package"
		nsbdcatch {file delete $nsbStoreName}
	    }
	}

	alwaysEvalFor "error in final installation, inconsistent package!" {} {
	    installNupPaths
	}
	
	if {([info exists nupContents(paths)] && ($nupContents(paths) != "")) ||
	    ([info exists nupContents(removePaths)] &&
					($nupContents(removePaths) != ""))} {
	    # only do postUpdateCommands if there was actually a change
	    executeUpdateCommands $package "postUpdateCommands" nupContents 1
	}

	if {$anotherInstallTop == ""} {
	    set anotherInstallTop $top
	}
    }

    nrdCommitUpdates
}

#
# Find existing stored '.nsb' file for package, if any.
# If second parameter is non-zero, return list of all, otherwise return
#   just first one
#
proc findStoredNsbfile {package {all 0} {executableTypes "{*}"}} {
    if {$executableTypes == "{*}"} {
	set executableTypes [cleanSubsts [nrdLookup $package executableTypes]]
    }
    if {$executableTypes == ""} {
	global cfgContents
	if {[info exists cfgContents(executableTypes)]} {
	    set executableTypes $cfgContents(executableTypes)
	} else {
	    set executableTypes [list ""]
	}
    }
    set versions [cleanSubsts [nrdLookup $package versions]]
    if {$versions == ""} {
	set versions [list ""]
    }
    set nameList ""
    foreach executableType $executableTypes {
	foreach version $versions {
	    set nsbStoreName [getNsbStoreName $package $executableType $version]
	    if {[lsearch -exact $nameList $nsbStoreName] >= 0} {
		continue
	    }
	    if {[file exists $nsbStoreName]} {
		if {!$all} {
		    return $nsbStoreName
		}
		lappend nameList $nsbStoreName
	    }
	}
    }
    return $nameList
}

#
# Get rid of nasty characters that could interfere in a filename
#
proc cleanSubst {string} {
    regsub -all {[ /]} $string - string
    return $string
}

#
# Do the same for a list of strings
#
proc cleanSubsts {strings} {
    set newStrings ""
    foreach string $strings {
	lappend newStrings [cleanSubst $string]
    }
    return $newStrings

}

#
# Return valid executableTypes for this package
# If there are registered executableTypes, and the package only contains
#   other types, raise an error
#
proc getExecutableTypes {package nsbContentsName} {
    upvar $nsbContentsName contents
    set types [nrdLookup $package executableTypes]
    global cfgContents
    if {($types == "") && [info exists cfgContents(executableTypes)]} {
	set types $cfgContents(executableTypes)
    }
    set executableTypes ""
    if {[info exists contents(executableTypes)]} {
	# intersect the .nsb's executableTypes with the registered ones
	set subtypes $contents(executableTypes)
	if {$types == ""} {
	    # none were in config file or registry, accept them all
	    set types $subtypes
	}
	foreach t $types {
	    set idx [lsearch -glob $subtypes $t]
	    if {$idx >= 0} {
		lappend executableTypes [lindex $subtypes $idx]
	    }
	}
	if {($types != "") && ($executableTypes == "")} {
	    nsbderror "executableType(s) do not match any of those registered for package $package"
	}
    }
    return $executableTypes
}

#
# Raise an error if required operating system versions are not supported
#   for any executableTypes in the list
#
proc checkOSVersions {executableTypes} {
    global cfgContents nsbContents
    foreach et $executableTypes {
	set min [list executableTypes $et minOSVersion]
	set max [list executableTypes $et maxOSVersion]
	if {[info exists cfgContents($max)] && [info exists nsbContents($min)]} {
	     set have $cfgContents($max)
	     set want $nsbContents($min)
	     if {[compareVersions $have $want] < 0} {
		nsbderror "$et OS version $have too old, must be at least $want"
	    }
	}
	if {[info exists cfgContents($min)] && [info exists nsbContents($max)]} {
	     set have $cfgContents($min)
	     set want $nsbContents($max)
	     if {[compareVersions $have $want] > 0} {
		nsbderror "$et OS version $have too new, must be at most $want"
	    }
	}
    }
}

#
# Return list of required packages from nsbContents that are not installed or
#     not the right version.  For each package, return three items on the list:
#   1. package name
#   2. string reason of why it doesn't meet the requirement
#   3. A substituted url of the 'nsb' files for the package, if supplied
#	and if the package was not registered
#
proc unmetRequirementPackages {executableTypes} {
    global nsbContents
    set rp "requiredPackages"
    if {![info exists nsbContents($rp)]} {
	return ""
    }
    set ans ""
    foreach p $nsbContents($rp) {
	if {![nrdPackageRegistered $package]} {
	    set nsbUrl ""
	    if {[info exists nsbContents([list $rp $p nsbUrl])]} {
		set nsbUrl [lindex [substituteNsbUrl \
		  $nsbContents([list $rp $p nsbUrl]) $p $executableTypes ""] 0]
	    }
	    lappend ans $p "not registered" $nsbUrl
	    continue
	}
	set versions [nrdLookup $p "versions"]
	if {[info exists nsbContents([list $rp $p minVersion])]} {
	    set minVersion $nsbContents([list $rp $p minVersion])
	    set ok 0
	    foreach version $versions {
		if {[compareVersions $version $minVersion] >= 0} {
		    set ok 1
		    break
		}
	    }
	    if {!$ok} {
		lappend ans $p "version too old, wanted $minVersion" ""
		continue
	    }
	}
	if {[info exists nsbContents([list $rp $p maxVersion])]} {
	    set maxVersion $nsbContents([list $rp $p maxVersion])
	    set ok 0
	    foreach version $versions {
		if {[compareVersions $version $maxVersion] <= 0} {
		    set ok 1
		    break
		}
	    }
	    if {!$ok} {
		lappend ans $p "version too new, wanted $maxVersion" ""
		continue
	    }
	}
	set minGeneratedAt ""
	set maxGeneratedAt ""
	if {[info exists nsbContents([list $rp $p minGeneratedAt])]} {
	    set minGeneratedAt $nsbContents([list $rp $p minGeneratedAt])
	}
	if {[info exists nsbContents([list $rp $p maxGeneratedAt])]} {
	    set maxGeneratedAt $nsbContents([list $rp $p maxGeneratedAt])
	}
	if {($minGeneratedAt != "") || ($maxGeneratedAt != "")} {
	    foreach nsbStoreFile [findStoredNsbfile $p 1 $executableTypes] {
		catch {unset tmpContents}
		nsbdParseFile $nsbStoreFile "nsb" tmpContents
		set generatedAt $tmpContents(generatedAt)
		if {$minGeneratedAt != ""} {
		    if {[compareTimes $generatedAt $minGeneratedAt] >= 0} {
			set minGeneratedAt ""
			if {$maxGeneratedAt == ""} {
			    break
			}
		    }
		}
		if {$maxGeneratedAt != ""} {
		    if {[compareTimes $generatedAt $maxGeneratedAt] <= 0} {
			set maxGeneratedAt ""
			if {$minGeneratedAt == ""} {
			    break
			}
		    }
		}
	    }
	}
	if {$minGeneratedAt != ""} {
	    lappend ans $p "generatedAt too old, wanted $minGeneratedAt" ""
	    continue
	}
	if {$maxGeneratedAt != ""} {
	    lappend ans $p "generatedAt too new, wanted $maxGeneratedAt" ""
	    continue
	}
    }
    return $ans
}

#
# Return the default installTop
#
proc defaultInstallTop {} {
    global cfgContents
    if {[info exists cfgContents(installTop)]} {
	return $cfgContents(installTop)
    }
    return "~/nsbd"
}

#
# Get the installTop value for this package using given executableType
# See the description of the installTop configuration keyword for details.
#
proc getInstallTop {package executableType version} {
    set top [defaultInstallTop]
    set top [file join $top [nrdLookup $package installTop]]
    set top [substituteStorePath installTop $top $package $executableType $version]

    # by not passing nsbContents, we don't apply substitutions from .nsb file
    # Don't need to pass an executableType or version because they were
    #   already done by substituteStorePath
    initPathSubstitutions $package bogusContents "" ""
    set subtop [lindex [substitutePath $top] 1]
    if {$subtop == ""} {
	nsbderror "installTop $top for package $package is deleted
    by pathSubs or regSubs in config file or registry database"
    }
    set top $subtop

    # canonialize the name a little bit for comparisons
    set top [cleanPath $top]
    if {[string index $top [expr [string length $top] - 1]] != "/"} {
	return $top/
    }
    return $top
}

#
# Return the default nsbStorePath
#
proc defaultNsbStorePath {} {
    global cfgContents
    if {[info exists cfgContents(nsbStorePath)]} {
	return $cfgContents(nsbStorePath)
    }
    set nsbdpath $cfgContents(nsbdpath)
    if {[string index $nsbdpath [expr [string length $nsbdpath] - 1]] == "/"} {
	return $nsbdpath
    }
    return $nsbdpath/
}

#
# Return the name of the file that the '.nsb' file will be stored under
#   for the given package, executableType and version.
# See the description of the nsbStorePath configuration keyword for details.
#
proc getNsbStoreName {package executableType version} {
    set storePath [nrdLookup $package nsbStorePath]
    if {$storePath == ""} {
	set storePath [defaultNsbStorePath]
    }
    set storePath [substituteStorePath nsbStorePath $storePath \
				$package $executableType $version]
    set top [getInstallTop $package $executableType $version]
    set storeDir [file join $top $storePath]
    if {([string index $storePath [expr [string length $storePath] - 1]] == "/")
		    || [file isdirectory $storeDir]} {
	return [file join $storeDir ${package}.nsb]
    }
    return ${storeDir}
}

#
# Substitute %P, %E, and %V if any in path.
#
proc substituteStorePath {var path package executableType version} {
    foreach {subst keyname} "P package E executableType V version" {
	if {[regexp "%$subst" $path]} {
	    set varval [set $keyname]
	    if {$varval == ""} {
		nsbderror "$var $path for package $package contains %$subst and $keyname unknown"
	    }
	    regsub -all "%$subst" $path $varval path
	}
    }
    return $path
}

#
# Substitute %P, %E, and %V if any in nsbUrl for given package,
#    executableTypes, and versions
# Does not assume that executableTypes and versions are "cleanSubst"ed already
# Returns list of substituted nsbUrls
#

proc substituteNsbUrl {nsbUrl package executableTypes versions} {
    regsub -all "%P" $nsbUrl $package nsbUrl
    set nsbUrls [list $nsbUrl]
    if {[regexp %V $nsbUrl]} {
	# add an extra, empty version
	lappend versions ""
	set nsbUrls [multiSubstitute %V $nsbUrls [cleanSubsts $versions]]
    }
    if {[regexp %E $nsbUrl]} {
	if {$executableTypes == ""} {
	    nsbderror "No executableTypes in package $package to substitute for %E in nsbUrl"
	}
	set nsbUrls [multiSubstitute %E $nsbUrls [cleanSubsts $executableTypes]]
    }
    return $nsbUrls
}

#
# verify that all files at $urls are identical to $localFile
# If one of them is not, return it, otherwise return empty.
#
proc compareUrls {localFile urls} {
    foreach url $urls {
	set scratchName [scratchAddName "ver"]
	urlCopy $url $scratchName
	if {![compareFiles $localFile $scratchName]} {
	    return $url
	}
	scratchClean $scratchName
    }
    return ""
}

#
# verify that all applicable URLs in nsbContents are accessible and
#    match checksum info
#
proc fetchAllNsbContents {executableTypes versions nsbStoreName} {
    global nsbContents nsbNupKeys

    set package $nsbContents(package)

    if {![info exists nsbContents(nsbUrl)]} {
	progressmsg "Note: no nsbUrl keyword, user may not know where to find '.nsb' file"
    } else {
	set differentUrl [compareUrls $nsbStoreName \
			    [substituteNsbUrl $nsbContents(nsbUrl) \
				$package $executableTypes $versions]]
	if {$differentUrl != ""} {
	    nsbderror "file at $differentUrl not identical"
	}
    }

    addKeys $nsbNupKeys nsbContents subContents nsb
    substitutePaths $package nsbContents subContents \
	    					$executableTypes $versions 2
    validatePaths subContents [calculateValidPaths subContents]

    if {![info exists nsbContents(paths)]} {
	progressmsg "No applicable paths in package"
    } else {
	set mdType $subContents(mdType)
	set pathsInfo ""
	set scratchName [scratchAddName "ver"]
	foreach path $nsbContents(paths) {
	    if {![info exists nsbContents([list paths $path length])]} {
		# must be directory or link, skip it
		continue
	    }
	    set expectedMdData [list $nsbContents([list paths $path length]) \
				    $nsbContents([list paths $path $mdType])]
	    lappend pathsInfo [list $path $scratchName "" $expectedMdData]
	}
	multiFetchPaths nsbContents $mdType $pathsInfo
	scratchClean $scratchName
    }
}

#
# Load the old nsb file for comparison if available
#
proc loadOldNsbContents {nsbStoreFile package} {
    global oldnsbContents
    if {[file exists $nsbStoreFile]} {
	# read old contents into oldnsbContents
	progressmsg "Loading $nsbStoreFile"
	nsbdParseFile $nsbStoreFile "nsb" oldnsbContents
	alwaysEvalFor $nsbStoreFile {} {
	    if {![info exists oldnsbContents(package)]} {
		nsbderror "no package keyword in file"
	    }
	    if {$oldnsbContents(package) != $package} {
		nsbderror "unexpected package: $oldnsbContents(package)"
	    }
	}
    } else {
	catch {unset oldnsbContents}
    }
}

#
# Do any updates that are to be made to the registry database
# Assumes paths have already been loaded into nupContents
#
proc nsbUpdateNrd {nsbScratch package executableTypes versions} {
    global cfgContents nsbContents oldnsbContents nupContents nrdKeytable
    global procNsbType

    if {$procNsbType == "preview"} {
	return
    }
    #
    # Update the registered versions
    #
    if {![info exists nupContents(paths)]} {
	# will be deleting empty package
	set versions ""
    }
    if {[string first "%V" [getNsbStoreName $package "%E" "%V"]] < 0} {
	# there is no %V in the nsbStorePath, so overwrite existing versions
	nrdUpdate $package "set" "versions" $versions
    } else {
	if {[info exists oldnsbContents(version)]} {
	    # NOTE: this check is historical and only defensive at this point:
	    #   since we know there's a %V in the nsbStoreName and only one
	    #   version is supported in the .nsb file, the old .nsb file
	    #   can only contain the same version
	    set oldVersion $oldnsbContents(version)
	    if {[lsearch -exact $versions $oldVersion] < 0} {
		# remove old versions that have been deleted
		nrdUpdate $package "delete" "versions" $oldVersion
	    }
	}
	# add any new versions
	foreach version $versions {
	    nrdUpdate $package "add" "versions" $version
	}
    }

    if {$procNsbType == "register-update"} {
	# all the values have already been selected when registering,
	#  but do look for conflicting validPaths
	findConflictingValidPaths $package [nrdLookup $package validPaths]
	return
    }

    if {[info exists cfgContents(maintainerUpdatableKeys)]} {
	set updatableKeys $cfgContents(maintainerUpdatableKeys)
	# I don't want to support changing installTop (hard to do, and not
	#   very useful), but I want to display a warning if it changes,
	#   so that's why it's in the loop below.  Make sure it's not
	#   in the updatableKeys list.
	if {[set idx [lsearch -exact $updatableKeys "installTop"]] >= 0} {
	    set updatableKeys [lreplace $updatableKeys $idx $idx]
	}
    } else {
	set updatableKeys "nsbUrl"
    }

    foreach keyname {nsbUrl maintainers installTop validPaths \
			preUpdateCommands postUpdateCommands} {
	set lookupVal [nrdLookup $package $keyname]
	if {[lsearch -exact $updatableKeys $keyname] < 0} {
	    set updatable 0
	    # for non-updatable keys, look up value in old .nsb file
	    #  instead of in the database, to warn only about changes
	    if {[info exists oldnsbContents($keyname)]} {
		set oldVal $oldnsbContents($keyname)
	    } else {
		set oldVal ""
	    }
	} else {
	    set updatable 1
	    set oldVal $lookupVal
	}
	if {$keyname == "validPaths"} {
	    set newVal [calculateValidPaths nupContents]
	} else {
	    if {![info exists nsbContents($keyname)]} {
		continue
	    }
	    set newVal $nsbContents($keyname)
	}
	if {$oldVal == $newVal} {
	    # no change
	    continue
	}
	if {$oldVal == ""} {
	    if {$newVal == $lookupVal} {
		# no old value, but new value is same as what was registered
		continue
	    }
	    set fromTo "\n  was not set before but is now set to"
	} else {
	    set fromTo " changed from"
	}
	if {$nrdKeytable([list packages $keyname])} {
	    # a list
	    if {$oldVal != ""} {
		foreach val $oldVal {
		    append fromTo "\n    $val"
		}
		append fromTo "\n  to"
	    }
	    foreach val $newVal {
		append fromTo "\n    $val"
	    }
	} else {
	    # not a list
	    if {$oldVal != ""} {
		append fromTo "\n    $oldVal\n  to\n    $newVal"
	    } else {
		append fromTo "\n    $newVal"
	    }
	}
	if {!$updatable} {
	    warnmsg "Warning: non-maintainerUpdatableKey \"$keyname\"$fromTo\n  in the new $package '.nsb' file; registry not updated"
	    continue
	}
	# do per key-type checking before accepting the change
	switch $keyname {
	    nsbUrl {
		# make sure that all new nsbUrls are accessible
		set differentUrl [compareUrls $nsbScratch \
				    [substituteNsbUrl $newVal \
					$package $executableTypes $versions]]
		if {$differentUrl != ""} {
		    nsbderror "file at $differentUrl not identical, won't update registered nsbUrl"
		}
	    }
	    maintainers {
		# DWD TO DO
		nsbderror "remotely updating registered maintainers not implemented yet"
	    }
	    validPaths {
		findConflictingValidPaths $package $newVal \
		    "cannot update maintainerUpdatableKey 'validPaths' because"
	    }
	}
	updatemsg "\"$keyname\" for package $package$fromTo"
	nrdUpdate $package "set" $keyname $newVal
    }

}

#
# Look for conflicting validPaths in a package other than $package
#  at the same installTop as $package.  If find a conflict, raise an
#  error, prepending optional second parameter to the message.
#
proc findConflictingValidPaths {package validPaths {msg ""}} {
    set executableType "UnLiKeLy"
    set version "uNlIkElY"
    set installTop [getInstallTop $package $executableType $version]

    # add asterisks to directories in validPaths
    set newValidPaths ""
    foreach path $validPaths {
	if {[string index $path [expr [string length $path] - 1]] == "/"} {
	    lappend newValidPaths ${path}*
	} else {
	    lappend newValidPaths $path
	}
    }
    set validPaths $newValidPaths

    foreach otherPackage [nrdPackages] {
	if {$otherPackage == $package} {
	    continue
	}
	if {$installTop != [getInstallTop $otherPackage \
					$executableType $version]} {
	    continue
	}
	set otherValidPaths [nrdLookup $otherPackage validPaths]
	foreach otherPath $otherValidPaths {
	    if {[string index $otherPath \
			[expr [string length $otherPath] - 1]] == "/"} {
		set otherPath ${otherPath}*
	    }
	    foreach path $validPaths {
		if {[string match $path $otherPath] ||
			[string match $otherPath $path]} {
		    if {$msg != ""} {
			append msg "\n  a "
		    } else {
			set msg "A "
		    }
		    append msg \
"validPath for package '$package'
    $path
  conflicts with a validPath in pre-registered package '$otherPackage'
    $otherPath
  under the same installTop"
		    nsbderror $msg
		}
	    }
	}
    }
}

#
# Process the already-loaded nsbContents for a single installTop
#   and create the nupContents.
# Return 0 if there were no paths in nsbContents after substitution,
#  otherwise return 1
#
proc makeNupContents {nsbStoreFile nsbScratch installTop \
					executableTypes versions} {
    global nsbContents oldnsbContents nupContents nsbNupKeys procNsbType

    if {$executableTypes != ""} {
	if {[llength $executableTypes] > 1} {
	    set plural "s"
	} else {
	    set plural ""
	}
	set extra "\n   for executableType$plural $executableTypes"
    } else {
	set extra ""
    }
    catch {unset nupContents}
    addKeys $nsbNupKeys nsbContents nupContents nsb

    set package $nupContents(package)
    set nupContents(installTop) $installTop
    if {$executableTypes != ""} {
	set nupContents(executableTypes) $executableTypes
    }
    if {$versions != ""} {
	set nupContents(version) [lindex $versions 0]
    }

    #
    # Look first for updates that may be needed to registry database
    #
    loadOldNsbContents $nsbStoreFile $package

    substitutePaths $package nsbContents nupContents \
				$executableTypes $versions

    nsbUpdateNrd $nsbScratch $package $executableTypes $versions


    progressmsg "Calculating updates to install in $installTop$extra"

    if {$procNsbType != "preview"} {
	#
	# make temporary directory and set the extra temporaryTop keyword
	#
	set temporaryTop [makeTemporaryTop $package $installTop]
	set nupContents(temporaryTop) $temporaryTop
    }

    #
    # determine which files need to be loaded
    #
    set substitutedPaths ""
    if {![info exists nsbContents(paths)]} {
	updatemsg "no paths keyword found"
    } else {
	validatePaths nupContents [nrdLookup $package "validPaths"]
	set mdType $nupContents(mdType)
	if {[info exists oldnsbContents(paths)]} {
	    alwaysEvalFor $nsbStoreFile {} {
		substitutePaths $package oldnsbContents oldnupContents \
				    $executableTypes $versions
	    }
	}
	set newpaths ""
	if {[info exists nupContents(paths)]} {
	    set substitutedPaths $nupContents(paths)
	}
	foreach path $substitutedPaths {
	    set pathsTable($path) ""
	    if {[info exists nupContents([list paths $path isDirectory])]} {
		# a directory
		if {[info exists oldnupContents([list paths $path isDirectory])]} {
		    unset nupContents([list paths $path isDirectory])
		    continue
		}
		# fall through
	    } elseif {[info exists nupContents([list paths $path linkTo])]} {
		# a link
		set link $nupContents([list paths $path linkTo])
		if {[info exists oldnupContents([list paths $path linkTo])]} {
		    set oldlink $oldnupContents([list paths $path linkTo])
		    if {$link == $oldlink} {
			# hasn't changed, no need to include it
			unset nupContents([list paths $path linkTo])
			continue
		    }
		}
		# fall through
	    } elseif {[info exists oldnupContents([list paths $path length])] &&
		    [info exists oldnupContents([list paths $path $mdType])] &&
			($nupContents([list paths $path length]) ==
			    $oldnupContents([list paths $path length])) &&
			($nupContents([list paths $path $mdType]) ==
				$oldnupContents([list paths $path $mdType])) &&
			($nupContents([list paths $path mode]) ==
				$oldnupContents([list paths $path mode]))} {
		# file is the same as the old one, don't need to update
		unset nupContents([list paths $path length])
		unset nupContents([list paths $path $mdType])
		unset nupContents([list paths $path mode])
		continue
	    }
	    lappend newpaths $path
	}
	set nupContents(paths) $newpaths
    }

    #
    # determine which files need to be removed.
    # Note that a removed directory does not necessarily mean
    #  that the directory will need to be removed if there are
    #  still files that use the directory.  They are sorted in
    #  decreasing order to make sure all files will be removed
    #  before their enclosing directories.
    #
    set removePaths ""
    if {[info exists oldnupContents(paths)]} {
	foreach path $oldnupContents(paths) {
	    if {![info exists pathsTable($path)]} {
		if {$path != ""} {
		    lappend removePaths $path
		}
	    }
	}
    }
    set nupContents(removePaths) [lsort -decreasing $removePaths]

    return [expr {$substitutedPaths != ""}]
}

#
# Make the temporaryTop directory, and return its name
#
proc makeTemporaryTop {package installTop} {
    global cfgContents
    if {[info exists cfgContents(installTmp)]} {
	set topdir [file join $installTop $cfgContents(installTmp)]
    } else {
	set topdir $installTop
    }
    set temporaryTop [file join $topdir "tmp[pid]"]
    addToScratchFiles $temporaryTop
    nsbdcatch {file mkdir $temporaryTop}
    return $temporaryTop
}

#
# Substitute all paths in the fromContents table and load them into
#   the paths keyword in the toContents table.  Also add the internal keyword
#   mdType with the message digest type into the toContents table.
# If substituteLevel is 0, do all normal substitutions.  If 1 or more, ignore 
#  duplicate paths. If 2, don't include local config and registry substitutions.
#
#
proc substitutePaths {package fromContentsName toContentsName \
				executableTypes versions {substituteLevel 0}} {
    upvar $fromContentsName fromContents
    upvar $toContentsName toContents

    if {![info exists fromContents(paths)]} {
	return
    }
    set ignoreDups 0
    set contentsOnly 0
    if {$substituteLevel > 0} {
	set ignoreDups 1
	if {$substituteLevel > 1} {
	    set contentsOnly 1
	}
    }
    initPathSubstitutions $package fromContents $executableTypes $versions \
		$contentsOnly
    set mdType ""
    set toContents(mdType) ""
    set newpaths ""
    global knownMdTypes
    foreach path $fromContents(paths) {
	if {[set msg [relativePathCheck $path]] != ""} {
	    nsbderror $msg
	}
	set ans [substitutePath $path]
	set nonPortable [lindex $ans 0]
	set sPath [lindex $ans 1]
	if {$sPath == ""} {
	    # not applicable to this update
	    continue
	}
	if {[set msg [relativePathCheck $sPath]] != ""} {
	    nsbderror $msg
	}
	if {$nonPortable} {
	    # the nonPortable keyword is temporary and internal
	    set toContents([list paths $sPath nonPortable]) ""
	}

	set lastc [string index $sPath [expr [string length $sPath] - 1]]
	if {$lastc == "/"} {
	    # the isDirectory keyword is temporary and internal
	    set toContents([list paths $sPath isDirectory]) ""
	    # fall through
	} elseif {[info exists fromContents([list paths $path linkTo])]} {
	    # a link
	    set link $fromContents([list paths $path linkTo])
	    if {[set msg [relativePathCheck $link]] != ""} {
		nsbderror $msg
	    }
	    set ans [substitutePath $fromContents([list paths $path linkTo])]
	    set nonPortable [lindex $ans 0]
	    set link [lindex $ans 1]
	    if {$link == ""} {
		# not applicable to this update
		continue
	    }
	    if {[set msg [relativePathCheck $link]] != ""} {
		nsbderror $msg
	    }
	    if {$nonPortable} {
		set toContents([list paths $sPath nonPortable]) ""
	    }
	    set toContents([list paths $sPath linkTo]) $link
	    # fall through
	} elseif {![info exists fromContents([list paths $path length])]} {
	    nsbderror "no length for file $path"
	} elseif {[if {$mdType == ""} {
			# figure out which message digest type to use
			foreach t $knownMdTypes {
			    if {[info exists fromContents([list paths $path $t])]} {
				set mdType $t
				set toContents(mdType) $mdType
				break
			    }
			}
		    }
		    expr {$mdType == ""}]} {
	    nsbderror "no message digest for file $path of one of $knownMdTypes"
	} elseif {![info exists fromContents([list paths $path $mdType])]} {
	    nsbderror "no message digest of type $mdType for file $path"
	} else {
	    # copy the length and message digest to toContents, even though
	    #  they are temporary and internal and not officially
	    #  part of the toContents
	    set toContents([list paths $sPath length]) \
		    		$fromContents([list paths $path length])
	    set toContents([list paths $sPath $mdType]) \
		    		$fromContents([list paths $path $mdType])
	    # same goes for mode, and insert a default if missing
	    if {[info exists fromContents([list paths $path mode])]} {
		set toContents([list paths $sPath mode]) \
				    $fromContents([list paths $path mode])
	    } else {
		# fill in default
		set toContents([list paths $sPath mode]) "rw"
	    }
	    # fall through
	}
	# Note: the "loadPath" keyword is temporary and internal,
	#   not defined in the nup fileType tables so it would never
	#   be written out to a file
	if {[info exists toContents([list paths $sPath loadPath])] &&
							!$ignoreDups} {
	    set prevPath $toContents([list paths $sPath loadPath])
	    nsbderror "both $prevPath and $path install at $sPath"
	}
	set toContents([list paths $sPath loadPath]) $path
	lappend newpaths $sPath
    }
    if {$newpaths != ""} {
	set toContents(paths) $newpaths
    }

    if {$package == "nsbd"} {
	# calculate these two substitutions here while substitutePath is
	#  active and pass them in two temporary internal keywords
	set toContents(nsbdBackupFromPath) \
		[lindex [substitutePath "bin/nsbd"] 1]
	set toContents(nsbdBackupToPath) \
		[lindex [substitutePath "lib/nsbd/oldnsbd"] 1]
    }
}

#
# Do loop-invariant calculating for substitutions to avoid doing them
#   on every path.  Assumes called from the same proc as substitutePath
#   by saving cross-routine data in an "upvar" rather than a global.
# To not include local config and registry substitutions, set contentsOnly
#   to non-zero.  To not include nsbContents substitutions, pass in an
#   empty array as the nsbContents.
#
proc initPathSubstitutions {package nsbContentsName executableTypes versions
					    {contentsOnly 0}} {
    upvar $nsbContentsName contents
    upvar PathSubstitutions pathSubstitutions

    if {$contentsOnly} {
	set otherversions ""
	set otherexecutableTypes ""
    } else {
	#
	# Calculate the otherexecutableTypes and otherversions variables to
	#   list all known executableTypes and versions that are not
	#   listed in executableTypes and versions, for use in deleting
	#   paths that are not applicable
	if {[info exists contents(executableTypes)]} {
	    lappend executableTypesLists $contents(executableTypes)
	}
	global cfgContents
	lappend executableTypesLists [nrdLookup $package executableTypes]
	if {[info exists cfgContents(executableTypes)]} {
	    lappend executableTypesLists $cfgContents(executableTypes)
	}
	# currently there are no other sources for versions, but go through
	#  the same motions in case there are some in the future (and because
	#  there were some in the past!)
	if {[info exists contents(version)]} {
	    lappend versionsLists [list $contents(version)]
	} else {
	    set versionsLists ""
	}

	foreach keyname {executableTypes versions} {
	    set all [set $keyname]
	    set otherIndex [llength $all]
	    foreach list [set ${keyname}Lists] {
		foreach item $list {
		    luniqueAppend all $item
		}
	    }
	    set other$keyname [lrange $all $otherIndex end]
	}
    }

    #
    # Calculate the substitutions and deletions to perform.
    # Order is very important.
    #
    set pathSubstitutions ""
    if {$contentsOnly} {
	set sources contents
    } else {
	set sources {contents nrdLookup cfgContents}
    }
    foreach source $sources {
	foreach keyname {pathSubs regSubs} {
	    if {$source == "nrdLookup"} {
		set subs [nrdLookup $package $keyname]
	    } else {
		set contname "${source}($keyname)"
		if {[info exists $contname]} {
		    set subs [set $contname]
		} else {
		    set subs ""
		}
	    }
	    if {$keyname == "regSubs"} {
		set isReg 1
	    } else {
		set isReg 0
	    }
	    set singularKey [string range $keyname 0 \
				[expr [string length $keyname] - 2]]
	    foreach sub $subs {
		# Treat the substitution as a string rather than a list
		#  because don't want TCL to interpret any special
		#  characters.  The first whitespace separates the
		#  match part from the substitute part.
		regexp "^(\[^ \t\]*)\[ \t\]*(.*)" $sub x matchPart subPart

		# substitute %P
		regsub -all "%P" $matchPart $package matchPart

		# substitute %E and %V
		# "kind" is set to 0 if there are no %E or %V substitutions,
		#   1 if it is a substitution, or 2 if it is a deletion
		set kind 0
		set matchParts [list $matchPart]
		set othermatchParts ""
		foreach {subst listname} {E executableTypes V versions} {
		    if {[regexp %$subst $matchPart]} {
			set kind 1
			if {$subst == "E" && ([set $listname] == "")} {
			    nsbderror "a substitution contains %E but there are no executableTypes selected"
			}
			if {$matchParts == ""} {
			    set matchParts [list $matchPart]
			}
			set matchParts [multiSubstitute %$subst \
					    $matchParts [set $listname]]

			if {$othermatchParts == ""} {
			    set othermatchParts [list $matchPart]
			}
			set othermatchParts [multiSubstitute %$subst \
					$othermatchParts [set other$listname]]
		    }
		}
		foreach match $matchParts {
		    if {$keyname == "pathSubs"} {
			set matchLen [string length $match]
		    } else {
			set matchLen 0
		    }
		    lappend pathSubstitutions $matchLen $kind $match $subPart
		    debugmsg "will apply $singularKey \"$match\" => \"$subPart\""
		}
		foreach match $othermatchParts {
		    if {$keyname == "pathSubs"} {
			set matchLen [string length $match]
		    } else {
			set matchLen 0
		    }
		    lappend pathSubstitutions $matchLen 2 $match ""
		    debugmsg "will apply $singularKey deletion \"$match\""
		}
	    }
	}
    }
}

#
# Apply substitutions to the path.
# A substitution may result in a path being deleted entirely,
#  in which case an empty string is returned.
# Otherwise, returns a list of two items:
#  1) a boolean which is set to 0 if the path is portable, 1 if not portable
#  2) the substituted path
# If no substitutions are applicable, return the path unchanged.
#  
proc substitutePath {path} {
    upvar PathSubstitutions pathSubstitutions

    set nonPortable 0
    foreach {matchLen kind matchPart subPart} $pathSubstitutions {
	if {$matchLen != 0} {
	    # a pathSub
	    set pathLen [string length $path]
	    set endIndex [expr $matchLen - 1]
	    if {(($pathLen == $matchLen) && ($matchPart == $path)) ||
		 (($pathLen > $matchLen) && 
		    ([string range $path 0 $endIndex] == "$matchPart") &&
			(([string index $path $matchLen] == "/") ||
			    ([string index $matchPart $endIndex] == "/")))} {
		if {$kind == 2} {
		    return ""
		}
		if {$kind == 1} {
		    set nonPortable 1
		}
		set path "$subPart[string range $path $matchLen end]"
	    }
	} else {
	    # a regSub
	    if {$kind == 2} {
		# a deletion
		if {[regexp -- $matchPart $path]} {
		    return ""
		}
	    } else {
		set matched [regsub -all -- $matchPart $path $subPart path]
		if {($kind == 1) && $matched} {
		    set nonPortable 1
		}
	    }
	}
    }
    return [list $nonPortable $path]
}

#
# Calculate which paths should be considered valid for the
#   package in the contentsName table: validPaths are any
#   directory explicitly included plus any files that are not under
#   one of those directories.  Assumes substitutePaths has already
#   been called in the paths in contentsName.
#
proc calculateValidPaths {contentsName} {
    upvar $contentsName contents
    if {[info exists contents(validPaths)]} {
	return $contents(validPaths)
    }
    if {![info exists contents(paths)]} {
	return ""
    }
    set validPaths ""
    foreach path $contents(paths) {
	if {[info exists contents([list paths $path isDirectory])]} {
	    # a directory.  Remove any existing paths that are under
	    #  $path.  This is defensive coding, since normally the
	    #  directory will come before the paths anyway.
	    set newPaths ""
	    foreach validPath $validPaths {
		if {[string first $path $validPath] != 0} {
		    lappend newPaths $validPath
		}
	    }
	    set validPaths $newPaths
	}
	set gotmatch 0
	foreach validPath $validPaths {
	    if {$validPath == $path || 
		(([string first $validPath $path] == 0) &&
	         [info exists contents([list paths $validPath isDirectory])])} {
		set gotmatch 1
		continue
	    }
	}
	if {!$gotmatch} {
	    lappend validPaths $path
	}
    }
    return $validPaths
}

#
# validate that the path is valid for validPaths
# If it is not, raise an error
#
proc validatePath {path validPaths} {
    foreach vp $validPaths {
	if {[string index $vp [expr [string length $vp] - 1]] == "/"} {
	    set vp ${vp}*
	}
	if {[string match $vp $path]} {
	    return
	}
    }
    nsbderror "$path does not match any of the validPaths"
}

#
# Validate that all paths in contents table are within validPaths
# Raise an error if one doesn't match
#
proc validatePaths {contentsName validPaths} {
    upvar $contentsName contents
    global procNsbType
    if {($validPaths == "") && ($procNsbType == "preview")} {
	return
    }
    if {![info exists contents(paths)]} {
	return
    }
    if {($validPaths == "*") || ($validPaths == ".")} {
	# checking '*' here is just a shortcut
	# checking '.' here is obsolete but for backward compatibility
	return
    }
    foreach path $contents(paths) {
	validatePath $path $validPaths
	if {[info exists contents([list paths $path linkTo])]} {
	    validatePath $contents([list paths $path linkTo]) $validPaths
	}
    }
}

#
# Preview all the paths listed in nupContents 
#
proc previewNupPaths {} {
    global nupContents

    set package $nupContents(package)
    set installTop $nupContents(installTop)

    set firstmsg "Install top $installTop"
    foreach path $nupContents(paths) {
	set toPath [file join $installTop $path]
	set lastc [string index $path [expr [string length $path] - 1]]
	if {$lastc == "/"} {
	    if {![file exists $toPath] || \
				([file type $toPath] != "directory")} {
		previewNupUpdatemsg "Would make directory $path"
	    }
	} elseif {[info exists nupContents([list paths $path linkTo])]} {
	    previewNupUpdatemsg "Would link $path to $link"
	} else {
	    previewNupUpdatemsg "Would install $path"
	}
    }
    foreach path $nupContents(removePaths) {
	set lastc [string index $path [expr [string length $path] - 1]]
	if {$lastc == "/"} {
	    if {[file exists $toPath] && ([file type $toPath] == "directory")} {
		previewNupUpdatemsg "Would delete $path if empty"
	    }
	} else {
	    # Don't use "file isdirectory" in second comparison because then
	    #   links to directories appear as directories instead of links
	    if {[file isdirectory [file dirname $toPath]] &&
			[file exists $toPath] &&
				([file type $toPath] != "directory")} {
		previewNupUpdatemsg "Would delete $path"
	    }
	}
    }
    if {$firstmsg != ""} {
	progressmsg "No updates for files in package"
    }

}

proc previewNupUpdatemsg {msg} {
    uplevel {
	if {$firstmsg != ""} {
	    progressmsg $firstmsg
	    set firstmsg ""
	}
    }
    progressmsg $msg
}


#
# Return a path to $link relative to $relativePath.
#  $link is given as being relative to the top level.  
#  First, remove all common directory components.
#  Then, for each directory component remaining in
#  relativePath, prepend a "../".
#
proc relativeLinkPath {link relativePath} {
    set splitLink [split $link "/"]
    set splitPath [split $relativePath "/"]
    set endLink [expr [llength $splitLink] - 1]
    set endPath [expr [llength $splitPath] - 1]
    set idx 0
    while {($idx < $endLink) &&
		([lindex $splitLink $idx] == [lindex $splitPath $idx])} {
	incr idx
    }
    set link [join [lrange $splitLink $idx end] "/"]
    if {$link == "."} {
	set link ""
    }
    set dots ""
    while {$idx < $endPath} {
	append dots "../"
	incr idx
    }
    set link "$dots$link"
    if {$link == ""} {
	return "."
    }
    return $link
}

#
# load all the paths listed in nupContents into the temporaryTop which
#  is also listed in nupContents.  $anotherInstallTop, if not-empty is the
#  path of another instance of an installation with the same '.nsb' file,
#  but with a possibly different executableType or version; can use that
#  other directory to quickly copy in portable files instead of getting them
#  from the urls.
#
proc fetchNupPaths {anotherInstallTop} {
    global nupContents

    set package $nupContents(package)
    set installTop $nupContents(installTop)
    set temporaryTop $nupContents(temporaryTop)
    set pathsInfo ""
    set mdType $nupContents(mdType)

    foreach path $nupContents(paths) {
	set toPath [file join $temporaryTop $path]
	if {[info exists nupContents([list paths $path isDirectory])]} {
	    unset nupContents([list paths $path isDirectory])
	    nsbdcatch {file mkdir $toPath}
	} elseif {[info exists nupContents([list paths $path linkTo])]} {
	    nsbdcatch {file mkdir [file dirname $toPath]}
	    nsbdcatch {symLink [relativeLinkPath \
		$nupContents([list paths $path linkTo]) $path] $toPath}
	} else {
	    set mode $nupContents([list paths $path mode])
	    set perm 0
	    foreach {letter adder} {r 4 w 2 x 1} {
		if {[string first $letter $mode] >= 0} {
		    incr perm $adder
		}
	    }
	    set mode "0$perm$perm$perm"
	    set expectedMdData [list $nupContents([list paths $path length]) \
				    $nupContents([list paths $path $mdType])]
	    nsbdcatch {file mkdir [file dirname $toPath]}
	    if {($anotherInstallTop != "") && \
		   ![info exists nupContents([list paths $path nonPortable])]} {
		updatemsg "Copying $path from $anotherInstallTop"
		withOpen fd1 [file join $anotherInstallTop $path] "r" {
		    fconfigure $fd1 -translation binary
		    set fd2 [nsbdcatch {open $toPath "w" $mode}]
		    alwaysEvalFor $toPath {close $fd2} {
			fconfigure $fd2 -translation binary
			set mdData [$mdType -copychan $fd2 -chan $fd1]
		    }
		}
		compareMdDataFor $path $expectedMdData $mdData
	    } else {
		if {[info exists nupContents([list paths $path nonPortable])]} {
		    unset nupContents([list paths $path nonPortable])
		}
		set fromPath $nupContents([list paths $path loadPath])
		lappend pathsInfo \
		    [list $fromPath $toPath $mode $expectedMdData]
	    }
	    unset nupContents([list paths $path loadPath])
	    unset nupContents([list paths $path length])
	    unset nupContents([list paths $path $mdType])
	    unset nupContents([list paths $path mode])
	}
    }

    multiFetchPaths nupContents $mdType $pathsInfo

    catch {unset nupContents(topUrl)}
    catch {unset nupContents(multigetUrl)}
}

#
# If the multigetUrl keyword is available, do a multiget on
#   the paths in pathsInfo, otherwise get each one separately
#
proc multiFetchPaths {contentsName mdType pathsInfo} {
    upvar $contentsName contents
    if {[info exists contents(multigetUrl)]} {
	urlMultiMdCopy $contents(multigetUrl) $mdType $pathsInfo
	return
    }
    if {![info exists contents(topUrl)]} {
	nsbderror "topUrl and multigetUrl keyword both missing"
    }
    set topUrl $contents(topUrl)

    set endTop [expr [string length $topUrl] - 1]
    if {[string index $topUrl $endTop] == "/"} {
	set topUrl [string range $topUrl 0 [expr $endTop - 1]]
    }

    foreach pathInfo $pathsInfo {
	foreach {fromPath toPath mode expectedMdData} $pathInfo {}
	set mdData [urlMdCopy $topUrl/$fromPath $mdType $toPath $mode]
	compareMdDataFor $fromPath $expectedMdData $mdData
    }
}

#
# install all the paths listed in nupContents from the temporaryTop directory
#  to the installTop directory, and clean out the temporaryTop directory.
#
proc installNupPaths {} {
    global nupContents

    set temporaryTop $nupContents(temporaryTop)
    set installTop $nupContents(installTop)

    set firstmsg "Install top $installTop"
    if {$nupContents(package) == "nsbd"} {
	# special case to back up nsbd in case the update broke further updates
	set fromPath $nupContents(nsbdBackupFromPath)
	set toPath $nupContents(nsbdBackupToPath)
	set fullFromPath [file join $installTop $fromPath]
	set fullToPath [file join $installTop $toPath]
	if {($fromPath != "") && ($toPath != "") &&
	    ([lsearch $nupContents(paths) $fromPath] >= 0) &&
		[file readable $fullFromPath] &&
		    [file isdirectory [file dirname $fullToPath]]} {
	    installNupUpdatemsg "Making backup copy of $fromPath in $toPath"
	    nsbdcatch {file delete -force $fullToPath}
	    nsbdcatch {file copy $fullFromPath $fullToPath}
	}
    }
    foreach path $nupContents(paths) {
	set toPath [file join $installTop $path]
	set lastc [string index $path [expr [string length $path] - 1]]
	if {$lastc == "/"} {
	    if {![file exists $toPath] || \
				([file type $toPath] != "directory")} {
		installNupUpdatemsg "Making directory $path"
		nsbdcatch {file delete $toPath}
		nsbdcatch {file mkdir $toPath}
	    }
	} elseif {[info exists nupContents([list paths $path linkTo])]} {
	    set link [relativeLinkPath $nupContents([list paths $path linkTo]) $path]
	    installNupUpdatemsg "Linking $path to $link"
	    nsbdcatch {file mkdir [file dirname $toPath]}
	    nsbdcatch {file delete -force $toPath}
	    nsbdcatch {symLink $link $toPath}
	} else {
	    installNupUpdatemsg "Installing $path"
	    nsbdcatch {file mkdir [file dirname $toPath]}
	    nsbdcatch {file delete -force $toPath}
	    nsbdcatch {file rename [file join $temporaryTop $path] $toPath}
	}
    }
    scratchClean $temporaryTop
    unset nupContents(temporaryTop)
    foreach path $nupContents(removePaths) {
	set toPath [file join $installTop $path]
	set lastc [string index $path [expr [string length $path] - 1]]
	if {$lastc == "/"} {
	    # Do *NOT* use -force!!  Fail silently if the directory
	    #  is not empty
	    if {[file exists $toPath] && ([file type $toPath] == "directory")} {
		installNupUpdatemsg "Deleting $path if empty"
		catch {file delete $toPath}
	    }
	} else {
	    # Don't use "file isdirectory" in second comparison because then
	    #   links to directories appear as directories instead of links
	    if {[file isdirectory [file dirname $toPath]] &&
			[file exists $toPath] &&
				([file type $toPath] != "directory")} {
		installNupUpdatemsg "Deleting $path"
		set code [catch {file delete $toPath} string]
		if {$code > 0} {
		    updatemsg "problem deleting file, continuing: $string"
		}
	    }
	}
    }
    if {$firstmsg != ""} {
	updatemsg "No updates to files in package"
    }
}

proc installNupUpdatemsg {msg} {
    uplevel {
	if {$firstmsg != ""} {
	    updatemsg $firstmsg
	    set firstmsg ""
	}
    }
    updatemsg $msg
}

#
# Execute any updateCommands, passing an extra parameter of the name of
#  a file with an nsbdUpdate file with a dump of 
#
proc executeUpdateCommands {package commandsName contentsName {ignoreErrors 0}} {
    upvar $contentsName contents
    global cfgContents
    set singularName [string range $commandsName 0 \
			[expr [string length $commandsName] - 2]]
    set commands [nrdLookup $package $commandsName]
    if {[info exists cfgContents($commandsName)]} {
	set commands [concat $commands $cfgContents($commandsName)]
    }
    if {$commands == ""} {
	return
    }
    set scratchUpdName [scratchAddName "nup"]
    nsbdGenFile $scratchUpdName nup contents

    global env
    if {[info exists env(SHELL)]} {
	set shell $env(SHELL)
    } else {
	set shell sh
    }
    foreach command $commands {
	# use alwaysEvalFor just to add the command name to error messages
	alwaysEvalFor $command {} {
	    set command "$command $scratchUpdName"
	    updatemsg "Executing $singularName command:\n  $command"
	    set code [catch {exec $shell -c $command} result]
	    if {$code > 0} {
		if {$ignoreErrors} {
		    updatemsg "command had an error, ignoring:\n  $result"
		} else {
		    nsbderror "$result"
		}
	    } elseif {$result != ""} {
		updatemsg "command returned the following:\n  $result"
	    }
	}
    }
    scratchClean $scratchUpdName
}

#
# prompt for any unmet requiredPackages requirements
# Return 1 if registration can proceed.
# Return 0 if need to abort processing of registration.
#
proc promptUnmetRequirements {registerPackage executableTypes} {
    set unmet [unmetRequirementPackages $executableTypes]
    if {$unmet == ""} {
	return 1
    }
    set updateList ""
    set updateMsg ""
    set registerList ""
    set registerMsg ""
    set unavailableList ""
    set unavailableMsg ""
    foreach {package requirement nsbUrl} $unmet {
	if {$requirement == "not registered"} {
	    if {$nsbUrl == ""} {
		lappend unavailableList $package
	    } else {
		lappend registerList $package $nsbUrl
	    }
	} else {
	    lappend updateList $package $requirement
	}
    }
    if {$updateList != ""} {
	if {[llength $updateList] > 2} {
	    set thisPackage "these packages"
	    set msg "Warning: $thisPackage do not meet requirements:"
	    foreach {package requirement} $updateList {
		append msg "\n  $package $requirement"
	    }
	} else {
	    set package [lindex $updateList 0]
	    set thisPackage "package $package"
	    set msg "Warning: $thisPackage does not meet requirements."
	}
	append msg "\nDo you want to check for updates to $thisPackage"
	append msg "\n  instead of registeringing $registerPackage at this time?"
	set answer [tk_dialog .d "" $msg warning 0 \
	    "Check $thisPackage" "Skip check, proceed" "Cancel registration"]
	if {$answer == 2} {
	    return 0
	}
	if {$answer == 0} {
	    foreach {package requirement} $updateList {
		option-update $package
	    }
	    tk_dialog .d "" "Update check complete" "" 0 "Exit"
	    return 0
	}
    }
    if {$registerList != ""} {
	if {[llength $registerList] > 2} {
	    set thisPackage "these packages"
	    set msg "Warning: these required packages are not registered:"
	    foreach {package nsbUrl} $registerList {
		append msg "\n  $package"
	    }
	} else {
	    set package [lindex $registerList 0]
	    set thisPackage "package $package"
	    set msg "Warning: required $thisPackage is not registered."
	}
	append msg "\nDo you want to register $thisPackage instead of"
	append msg "\n  registering $registerPackage at this time?"
	set answer [tk_dialog .d "" $msg warning 0 "Register $thisPackage" \
		"Skip $thisPackage, proceed" "Cancel registrations"]
	if {$answer == 2} {
	    return 0
	}
	if {$answer == 0} {
	    foreach {package nsbUrl} $registerList {
		procfile $nsbUrl nsb
	    }
	    return 0
	}
    }
    if {$unavailableList != ""} {
	if {[llength $unavailableList] > 1} {
	    set thisPackage "these packages"
	    set msg "Warning: these required packages are not registered:"
	    foreach package $unavailableList {
		append msg "\n  $package"
	    }
	} else {
	    set package [lindex $unavailableList 0]
	    set thisPackage "package $package"
	    set msg "Warning: required $thisPackage is not registered."
	}
	append msg "\nNo information was provided about how to register"
	append msg "\n  for $thisPackage."
	append msg "\nDo you want to proceed with registering $registerPackage anyway?"
	set answer [tk_dialog .d "" $msg warning 0 \
			"Proceed with registration" "Cancel registration"]
	if {$answer == 1} {
	    return 0
	}
    }
    return 1
}

#
# prompt to register the package described in nsbContents
#

set promptKeys {
    nsbUrl maintainers minPollPeriod executableTypes installTop nsbStorePath
    pathSubs regSubs validPaths preUpdateCommands postUpdateCommands
}

proc promptToRegister {package executableTypes versions} {
    if {[nrdPackageRegistered $package]} {
	set answer [tk_dialog .d "" \
"The $package package is already registered.
  Do you want to overwrite it?" warning 0 "Proceed" "Cancel"]
	if {$answer == 1} {
	    return
	}
    }
	
    global nsbContents promptKeys promptValueScripts

    catch {unset promptValueScripts}

    set w ".register"
    toplevel $w
    wm withdraw $w
    wm title $w "NSBD registration of package $package"
    label $w.disclaimer -text \
{This is the Not-So-Bad Distribution (NSBD) package registration window.
IMPORTANT NOTE: a digital signature on a package only authenticates that the
package comes from the person who signed it.  It makes no guarantees on the
integrity of the person or package.  Register for packages at your own risk.    }
    pack $w.disclaimer
    global warnUpdateCommands
    set warnUpdateCommands 0
    foreach key $promptKeys {
	set numlines 1
	set commalist 0
	set defaultValue ""
	if {[info exists nsbContents($key)]} {
	    set value $nsbContents($key)
	} else {
	    set value [nrdLookup $package $key]
	}
	switch $key {
	    nsbUrl {
		set label "URL of '.nsb' file:  "
	    }
	    maintainers {
		set label "PGP ids of maintainers:  "
		set numlines 2
	    }
	    minPollPeriod {
		set label "Minimum poll period:  "
		if {$value == ""} {
		    set value "1d"
		    set defaultValue $value
		}
	    }
	    executableTypes {
		if {$value == ""} {
		    continue
		}
		set label "Executable types:  "
		set numlines [llength $executableTypes]
		if {$numlines < 2} {
		    set numlines 2
		} elseif {$numlines > 5} {
		    set numlines 5
		}
		set value $executableTypes
	    }
	    installTop {
		set label "Install directory:    "
		if {$value == ""} {
		    set value [defaultInstallTop]
		    set defaultValue $value
		}
	    }
	    nsbStorePath {
		set label "Path to store '.nsb' file: "
		if {$value == ""} {
		    set value [defaultNsbStorePath]
		    set defaultValue $value
		}
	    }
	    pathSubs {
		set label "User path substitutions:    "
		set numlines 2
		# don't use nsbContents for this one
		set value [nrdLookup $package $key]
	    }
	    regSubs {
		set label "User regexp substitutions: "
		set numlines 2
		# don't use nsbContents for this one
		set value [nrdLookup $package $key]
	    }
	    validPaths {
		set label "Valid paths:          "
		set numlines 5
		if {![info exists nsbContents($key)]} {
		    # if not manually in nsb file, make default be
		    #  calculated validPaths, taking priority over
		    #  any previous validPaths in the nrd
		    substitutePaths $package nsbContents subContents \
	    					$executableTypes $versions 1
		    set value [calculateValidPaths subContents]
		}
	    }
	    preUpdateCommands {
		if {$value != ""} {
		    set warnUpdateCommands 1
		}
		set label "Pre-update commands:   "
		set numlines 2
	    }
	    postUpdateCommands {
		if {$value != ""} {
		    set warnUpdateCommands 1
		}
		set label "Post-update commands: "
		set numlines 2
	    }
	}
	frame $w.${key}Frame
	pack $w.${key}Frame -fill x -padx 10 -pady 2
	set trimmedLabel [string trimright $label]
	button $w.${key}Label -text "$trimmedLabel" \
			-command "registryHelp $key" -takefocus 0
	pack $w.${key}Label -in $w.${key}Frame -side left
	set fill [string range $label [string length $trimmedLabel] end]
	label $w.${key}Fill -text "$fill" 
	pack $w.${key}Fill -in $w.${key}Frame -side left
	if {$numlines == 1} {
	    entry $w.$key -width 60
	    if {$commalist} {
		$w.$key insert end [join $value ","]
		set script "set data \[$w.$key get\]"
		append script {
		    regsub -all {[ \t]} $data "" data
		    split $data ","
		}
	    } else {
		$w.$key insert end $value
		set script "set v \[string trim \[$w.$key get\]\]"
		if {$defaultValue != ""} {
		    append script "\nexpr {\$v == \"$defaultValue\" ? \"\" : \$v}"
		}
	    }
	    pack $w.$key -in $w.${key}Frame -side left -expand 1 -fill x
	    bind $w.$key <Return> "event generate $w.$key <Tab>"
	} else {
	    text $w.$key -width 60 -height $numlines -wrap none \
			    -yscrollcommand "$w.${key}Scroll set"
	    scrollbar $w.${key}Scroll -command "$w.$key yview" -takefocus 0
	    $w.$key insert end [join $value "\n"]
	    pack $w.${key}Scroll -in $w.${key}Frame -side right -fill y
	    pack $w.$key -in $w.${key}Frame -side left -fill x -expand 1
#	    bind $w.$key <FocusIn> {
#		# This isn't exactly the same behavior as in an Entry,
#		#  but it is close enough
#		if {[%W compare insert == "end - 1 chars"]} {
#		    %W tag add sel 0.0 end
#		}
#	    }
	    set script "set data \[$w.$key get 0.0 end\]"
	    append script {
		set trimmedData ""
		foreach datum [split $data "\n"] {
		    set datum [string trim $datum]
		    if {$datum != ""} {
			lappend trimmedData $datum
		    }
		}
		set trimmedData
	    }
	}
	set promptValueScripts($key) $script
    }
    if {$warnUpdateCommands} {
	global updateCommandsApproved
	set updateCommandsApproved 0
	label $w.warnCommandsLabel -text \
{Pre or Post update commands were provided, which can be particularly dangerous.}
	pack $w.warnCommandsLabel
	checkbutton $w.warnCommandsButton -variable updateCommandsApproved -text \
{Check here to indicate that you have approved the update commands.}
	pack $w.warnCommandsButton
    }
    frame $w.line -bd 2 -height 4 -relief raised
    pack $w.line -fill x -padx 5
    frame $w.buttons
    pack $w.buttons -fill x
#   frame $w.buttons.default -relief sunken -bd 1
#   pack $w.buttons.default -side left -expand 1 -pady 4
#   frame $w.buttons.default.frame -relief raised -bd 1
#   pack $w.buttons.default.frame -padx 2 -pady 2
    button $w.attempt -text "Install and register" -command {installRegister}
    button $w.cancel -text "Cancel registration" -command "destroy $w"
    button $w.help -text "Help" -command {registryHelp overall}
#   pack $w.attempt -in $w.buttons.default.frame -padx 2 -pady 2
    pack $w.attempt -in $w.buttons -side left -expand 1 -pady 4
    pack $w.cancel -in $w.buttons -side left -expand 1 -pady 4
    pack $w.help -in $w.buttons -side left -expand 1 -pady 4
    foreach button "$w.attempt $w.cancel $w.help" {
	bind $button <Return> [$button cget -command]
    }
    tkCenter $w

    focus $w.nsbUrl
#    $w.nsbUrl selection range 0 end
#    $w.nsbUrl icursor end

    tkwait window $w
}

#
# Attempt installation of the package and register if successful
#
proc installRegister {} {
    set w ".register"
    global warnUpdateCommands updateCommandsApproved
    if {$warnUpdateCommands && !$updateCommandsApproved} {
	tk_dialog $w.d "" \
"You must check the box to approve
update commands before proceeding" warning 0 "Cancel"
	return
    }
    global ignoreSecurity cfgContents nsbContents promptKeys promptValueScripts
    set package $nsbContents(package)
    if {!$ignoreSecurity && ![info exists cfgContents(pgp)]} {
	set answer [tk_dialog $w.d "" \
"The 'pgp' configuration keyword is not set and the
  '-ignoreSecurity' command line option was not used.
  Do you want to proceed and ignore security?" \
				warning 0 "Proceed" "Cancel install"]
	if {$answer == 1} {
	    destroy $w
	    return
	}
	set ignoreSecurity 1
    }
    if {!$ignoreSecurity} {
	foreach id [eval $promptValueScripts(maintainers)] {
	    if {![pgpIDKnown $id]} {
		set keylist [list maintainers $id pgpKeyUrl]
		if {[info exists nsbContents($keylist)]} {
		    set answer [tk_dialog $w.d "" \
"The PGP public key of maintainer
    $id
is not known to PGP, but a URL for the key has
been supplied.  Do you want to register the key?" \
			    "" 0 "Register the key" "Cancel install"]
		    if {$answer == 1} {
			destroy $w
			return
		    }
		    set scratch [scratchAddName "key"]
		    urlCopy $nsbContents($keylist) $scratch
		    pgpAddKey $id $scratch
		    scratchClean $scratch
		} else {
		    nsbderror \
"The PGP public key of maintainer
    $id
is not registered and no pgpKeyUrl was supplied."
		}
	    }
	}
    }
    foreach key $promptKeys {
	if {[info exists promptValueScripts($key)]} {
	    nrdUpdate $package "set" $key [eval $promptValueScripts($key)]
	}
    }
    raiseMessagesWindow
    option-update $package "register-update"
    tk_dialog $w.d "" "Installation and registration\n  of $package complete" "" 0 "Exit"
    destroy $w
}

#
# Pop up a help window with info about the registry key $key
#
proc registryHelp {key} {
    if {$key == "overall"} {
	set msg \
"This is [nsbdVersion].
This window is for registering an NSBD-compliant package.

When you press 'Install and register', an attempt will
  be made to install the package.  If there are any errors
  in the installation, the registration will be backed out;
  if there are no errors, the package will be registered.
  You can later check for updates through command line
  options and won't need this window.

For help on individual registration items, click on the
  label for the item."
    } else {
	global nrdKeylist
	foreach {c k v} $nrdKeylist {
	    if {$k == "packages $key"} {
		set msg [join $c "\n"]
		break
	    }
	}
	if {$msg == ""} {
	    nsbderror "Internal error - cannot find comment for nrd key $key"
	}
    }
    tk_dialog .helpd "NSBD Help" $msg info 0 "Ok"
}
