#
# Handle command-line options 
# File created by Dave Dykstra <dwd@bell-labs.com>, 21 Nov 1996
#
# Copyright (C) 1996-7 by Dave Dykstra and Lucent Technologies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#  
# If those terms are not sufficient for you, contact the author to 
# discuss the possibility of an alternate license.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

#
# empty procedure for auto loading this file
#
proc ensure-options-loaded {} {
}

# Every third item in the list is the following:
#   1. A list of comment lines that apply to this option, suitable for
#	printed help about the option.  If the comment is empty, it will
#	not be included in help (used for obsolete commands).
#   2. A list of equivalent option names.  There should be corresponding tcl
#	functions that begin with "option".
#   3. A number with one of these possible values:
#	   0. Option takes no parameters.
#	   1. Option takes 1 parameter.  If not given on the command line,
#		option function will be invoked with an empty string.
#	   2. Option takes each remaining non-keyword setting argument
#		from the command line as a parameter.  If there were none,
#		the option function will be invoked once with an empty
#		parameter.
#	   3. Option takes the rest of the command line as a list parameter.
#
set infoOnlyOptions "-help -? -version -V -license"
set optionList {
 {{Print command-line help.}}
"-help -?" 1

 {{Print the current program version.}}
"-version -V" 0

 {{Print the license (GNU General Public License) and no-warranty details.}}
"-license" 0

 {{Register package described in the '.nsb' file at following filename or URL.}
  {Pops up a window to request verification of package registry information}
  {before registering and installing.  If a Graphical User Interface display}
  {is not available, register package by hand by editting 'registry.nrd' with}
  {information from the package's '.nsb' file and use '-update' on the package.}}
"-register" 1

 {}
"-installPackage -install" 2

 {{Check registered packages for updates and install the updates.  If package}
  {name is the special name "all", will check all registered packages.}}
"-update" 2

 {}
"-checkPackage -check" 2

 {{Same as -update except that only one package may be checked and no command}
  {line keyword settings are supported after the option; for use by remote}
  {invocation from a CGI script.}}
"-updateOnePackage" 3

 {{Same as -update, except only polls registered packages whose minPollPeriod}
  {has elapsed.  If package name is the special name "all", will poll all}
  {registered packages.}}
"-poll" 2

 {{Preview the following packages, either registered package names or '.nsb'}
  {files, displaying which files are ready to be installed or removed.  Note}
  {that if you preview a '.nsb' file for a package that is not registered,}
  {you will not be prompted for installation choices so defaults will be used.}}
"-preview" 2

 {}
"-verify" 2

 {{Fetch all the following packages, either registered package names or '.nsb'}
  {files, and verify that all files can be fetched and that the checksums match}
  {those that are in the '.nsb' files.  Each '.nsb' itself file will also be}
  {verified to be identical to the one referred to by the 'nsbUrl' keyword in}
  {the file.  This is useful for maintainers of packages to test their packages}
  {without installing anything.}}
"-fetchAll" 2

 {{Update the comment template in all following files.  Files can be any of}
  {the recognized types of "NSBD" format files, but PGP signatures will be}
  {deleted from '.nsb' files.  Existing comments and unrecognized keywords}
  {will be deleted, and comments will be inserted for every valid keyword.}
  {The original file will be saved with "old" added to the beginning of the}
  {name.}}
"-updateComments" 2

 {{Do not put a PGP signature when creating a '.nsb' file from a '.npd' file.}
  {WARNING: Use carefully, only when your users can trust the network over}
  {which they retrieve your '.nsb' file.  This must be before the '.npd' file}
  {on the command line.}}
"-unsigned" 0

 {{Ignore PGP signature when processing a '.nsb' file.  WARNING: Use carefully,}
  {only when you can trust the source of the '.nsb' file.  This must be before}
  {the -update or -poll option on the command line.}}
"-ignoreSecurity" 0

 {{Retrieve text at given url and send it to standard output.  This is a}
  {general purpose command that has very little to do with the rest of NSBD,}
  {it is just here for convenience, especially for invoking a remote CGI}
  {script, and because it just takes one extra line of code to implement.}}
"-getUrl" 1

 {{Send standard input as an http POST command to the given url and send the}
  {result to standard output.  As with -getUrl, this has very little to do}
  {with the rest of NSBD.}}
"-postUrl" 1
}
catch {unset optionTable}
foreach {c k v} $optionList {
    foreach s $k {
	set optionTable($s) $v
    }
}

#
# Command-line help 
#

proc option-? {arg} {
    if {$arg == ""} {
	usage
    } else {
	option-help $arg
    }
}

proc option-help {arg {fd stdout}} {
    global optionList fileTitles
    if {$arg == "options"} {
	puts $fd "These are the available command-line options:"
	set options ""
	foreach {c s v} $optionList {
	    foreach opt $s {
		if {$c != ""} {
		    lappend options $opt
		}
	    }
	}
	display3Columns $fd $options
	puts $fd {For more help on an individual option use "-help -option_name"}
	puts $fd {  or for more help on all options use "-help alloptions".}
    } elseif {$arg == "keywords"} {
	puts $fd \
{Below are the available command line keywords, set by key=value
If the value is a list, it should be separated by commas.
Keyword settings can appear before or after a filename on the
   command line, but if the same keyword appears both before and
   after a filename, the one before takes precedence for that file.}
	foreach fileType {nsb npd cfg} {
	    upvar #0 ${fileType}Keys keys
	    puts $fd "For $fileTitles($fileType) files:" 
	    display3Columns $fd $keys
	}
	puts $fd {For more help on an individual keyword use "-help keyword_name"}
	puts $fd {  or for more help on all keywords use "-help allkeywords".}
    } elseif {$arg == "synopsis"} {
	puts $fd "Synopsis of nsbd usage:\n"
	puts $fd [nsbdSynopsis]
    } elseif {$arg == "operation"} {
	puts $fd "Operation of nsbd:\n"
	puts $fd [nsbdOperation]
    } elseif {$arg == "security"} {
	puts $fd "Security model of nsbd:\n"
	puts $fd [nsbdOperation]
    } elseif {$arg == "gettingstarted"} {
	puts $fd "Getting started with nsbd:\n"
	puts $fd [nsbdGettingStarted]
    } elseif {$arg == "fileformat"} {
	ensure-nsbdformat-loaded
	global nsbdFormatSpec
	puts $fd $nsbdFormatSpec
    } elseif {$arg == "examples"} {
	puts $fd "Examples of using nsbd:\n"
	puts $fd [nsbdExamples]
    } else {
	set gotone 0
	foreach fileType {nsb npd cfg nrd} {
	    upvar #0 ${fileType}Keylist keylist
	    if {$arg == "allkeywords"} {
		if {$fileType == "nrd"} {
		    continue
		}
		puts $fd "For $fileTitles($fileType) files:" 
	    }
	    foreach {c k v} $keylist {
		set lastk [lindex $k end]
		if {$lastk == $arg} {
		    puts $fd "For $fileTitles($fileType) files:" 
		}
		if {($lastk == $arg) || 
			(($arg == "allkeywords") && ([llength $k] == 1))} {
		    set gotone 1
		    set indent "  "
		    foreach word $k {
			puts $fd "$indent$word"
			set indent "$indent  "
		    }
		    if {$fileType == "npd"} {
			global nsbKeylist 
			foreach {c2 k2 v2} $nsbKeylist {
			    if {($k == $k2) && ($c == $c2)} {
				set c [list "Same as in $fileTitles(nsb) file."]
				break
			    }
			}
		    }
		    foreach l $c {
			puts $fd "$indent$l"
		    }
		}
	    }
	}
	foreach {c s v} $optionList {
	    if {([lsearch -exact $s $arg] >= 0) || ($arg == "alloptions")} {
		if {$c == ""} {
		    continue
		}
		set gotone 1
		puts $fd [join $s " or "]
		foreach l $c {
		    puts $fd "  $l"
		}
	    }
	}
	if {!$gotone} {
puts $fd \
{Please specify one of the following parameters to the -help option:
  "synopsis" - for a synopsis of nsbd usage.
  "operation" - for a general description of the operation of nsbd.
  "security" - for help on the security model of nsbd.
  "gettingstarted" - for help on getting started with nsbd.
  "options" - for a list of the available command line options.
  "alloptions" - for help on all available options.
  -option_name - for help on a specific option name.
  "keywords" - for a list of available command line keywords.
  "allkeywords" - for help on all available command line keywords.
  keyword_name - for help on a specific command line keyword, or any keyword
		    or subkeyword inside one of the four types of nsbd files.
  "fileformat" - for a detailed description of the format of nsbd files.
  "examples" - for examples of nsbd usage and files.}
	}
    }
    nsbdExit
}

#
# Display the items onto $fd in 3 columns, indented 4 spaces
#
proc display3Columns {fd items} {
    set is ""
    foreach i [lsort $items] {
	if {[llength $i] == 1} {
	    lappend is $i
	    if {[llength $is] == 3} {
		puts $fd [string trimright \
		    [format "    %-25s%-25s%-25s" [lindex $is 0] \
			[lindex $is 1] [lindex $is 2]]]
		set is ""
	    }
	}
    }
    if {[llength $is] > 0} {
	puts $fd [string trimright \
		    [format "    %-25s%-25s%-25s" [lindex $is 0] \
			[lindex $is 1] [lindex $is 2]]]
    }
}

#
# Print version number
#

proc option-V {} {
    option-version
}

proc option-version {} {
    puts [nsbdVersion]
    puts [nsbdCopyright]
    nsbdExit
}

#
# Print the license
#

proc option-license {} {
    puts [getLicense]
    nsbdExit
}

#
# Process -register option
#

proc option-register {nsbfile} {
    global procNsbType
    if {$nsbfile == ""} {
	nsbderror "'.nsb' file or URL missing for -register option"
    }
    set procNsbType "register"
    set code [catch {startGui} string]
    if {$code == 1} {
	global cfgContents
	puts stderr \
"nsbd: Couldn't start Graphical User Interface; you can manually register
       for packages without one.  See nsbd -help gettingstarted."
	global errorInfo errorCode
	return -code $code -errorinfo $errorInfo -errorcode $errorCode $string
    }
    procfile $nsbfile "nsb"
}

proc option-installPackage {package} {
    option-install $package
}

proc option-install {package} {
    set msg "The -install option has changed to -register"
    append msg "\n(you may need to change your ~/.mailcap and reload in"
    append msg "\n Netscape by bring up Options->General Preferences)"
    nsbderror $msg
}

#
# do procfile on nsbUrls from a registered package
#
proc procnsbpackage {package} {
    set nsbUrl [getCmdkey nsbUrl]
    if {$nsbUrl == ""} {
	set nsbUrl [nrdLookup $package nsbUrl]
    }
    if {$nsbUrl == ""} {
	nsbderror "No nsbUrl registered for package $package"
    }
    foreach nsbUrl [substituteNsbUrl $nsbUrl $package \
			[nrdLookup $package executableTypes] \
			[nrdLookup $package versions]] {
	procfile $nsbUrl "nsb" $package
    }
}

#
# Process -update option
#

proc option-update {package {nsbType update}} {
    global procNsbType
    set procNsbType $nsbType
    if {$package == ""} {
	nsbderror {package name or "all" missing for -update option}
    }
    if {$package == "all"} {
	set packages [nrdPackages]
    } else {
	set packages [list $package]
    }
    foreach package $packages {
	validatePackageName $package
	procnsbpackage $package
    }
}

proc option-updateOnePackage {package} {
    if {$package == ""} {
	nsbderror "Package name not given"
    }
    option-update $package
}

proc option-checkPackage {package} {
    option-check $package
}

proc option-check {package} {
    nsbderror "The -check option has changed to -update"
}

#
# Process -poll option
#
proc option-poll {package} {
    global cfgContents
    if {$package == ""} {
	nsbderror {package name or "all" missing for -poll option}
    }
    if {$package == "all"} {
	set packages [nrdPackages]
    } else {
	set packages [list $package]
    }
    foreach package $packages {
	set minPollPeriod [nrdLookup $package "minPollPeriod"]
	if {$minPollPeriod == ""} {
	    if {[info exists cfgContents(minPollPeriod)]} {
		set minPollPeriod $cfgContents(minPollPeriod)
	    } else {
		set minPollPeriod "1d"
	    }
	}
	set number 0
	set unit "d"
	scan $minPollPeriod "%d%s" number unit
	if {$number == 0} {
	    progressmsg "Package $package is never polled"
	    continue
	}
	set lastTimePolled [nrdLookup $package lastTimePolled]
	set now [currentTime]
	if {$lastTimePolled != ""} {
	    set multiplier 0
	    switch -regexp $unit {
		m|M {set multiplier 60}
		h|H {set multiplier [expr {60 * 60}]}
		d|D {set multiplier [expr {60 * 60 * 24}]}
		w|W {set multiplier [expr {60 * 60 * 24 * 7}]}
		default {
		    nsbderror \
"minPollPeriod for package $package ($minPollPeriod)
  must have units of m, h, d, or w"
		}
	    }
	    set number [expr {round($number * $multiplier * 0.98)}]
	    set nextTime [httpTime [expr {[scanAnyTime $lastTimePolled] + $number}]]
	    if {[compareTimes $now $nextTime] < 0} {
		updatemsg "Poll time for package $package not yet arrived: $nextTime"
		continue
	    }
	}
	nrdUpdate $package set lastTimePolled $now
	set code [catch {option-update $package poll} message]
	if {$code > 0} {
	    global errorInfo errorCode
	    if {$errorCode != "NSBD INTERNAL"} {
		error $message $errorInfo $errorCode
	    }
	    #
	    # else continue to poll the next package
	    #
	    errormsg "Error polling package $package:\n  $message"
	    nrdAbortUpdates
	} else {
	    # in case it hadn't been done already; no effect if it had
	    nrdCommitUpdates
	}
    }
}

#
# Process -preview option
#

proc option-preview {arg} {
    global procNsbType
    if {$arg == ""} {
	nsbderror "'.nsb' file or URL or package name missing for -preview option"
    }
    set procNsbType "preview"
    if {[nrdPackageRegistered $arg]} {
	procnsbpackage $arg
    } else {
	procfile $arg "nsb"
    }
}

#
# Process -fetchAll option
#

proc option-fetchAll {arg} {
    global procNsbType
    if {$arg == ""} {
	nsbderror "'.nsb' file or URL or package name missing for -fetchAll option"
    }
    set procNsbType "fetchAll"
    if {[nrdPackageRegistered $arg]} {
	procnsbpackage $arg
    } else {
	procfile $arg "nsb"
    }
}

proc option-verify {arg} {
    nsbderror "The -verify option has changed to -fetchAll"
}


#
# Process -ignoreSecurity option
#
proc option-ignoreSecurity {} {
    global ignoreSecurity
    set ignoreSecurity 1
}

#
# Process -unsigned option
#
proc option-unsigned {} {
    global unsignedNsbfiles
    set unsignedNsbfiles 1
}

#
# Process -updateComments option
#
proc option-updateComments {fileName} {
    if {$fileName == ""} {
	usage {-updateComments requires a file name parameter}
    }
    progressmsg "Updating comments in $fileName"
    set fileType [lindex [nsbdParseFile $fileName "" contents] 0]
    set scratchName [scratchAddName "tem"]
    withOpen fd $scratchName "w" {
	nsbdGenTemplate $fd $fileType contents
    }
    set oldName [addFilePrefix $fileName old]
    nsbdcatch {file rename -force $fileName $oldName}
    nsbdcatch {file rename $scratchName $fileName}
    progressmsg "Old file saved in $oldName"
}

#
# Process -getUrl option
#
proc option-getUrl {url} {
    urlCopy $url -
}

#
# Process -postUrl option
#
proc option-postUrl {url} {
    urlCopy $url - "" "-query [read stdin]"
}
