#
# Interface to PGP for NSBD
# File created by Dave Dykstra <dwd@bell-labs.com>, 13 Dec 1996
#
#
# Copyright (C) 1996-7 by Dave Dykstra and Lucent Technologies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#  
# If those terms are not sufficient for you, contact the author to 
# discuss the possibility of an alternate license.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

#
# Get the name of the pgp program, and set up any environment
#  variables it needs
#
proc getpgp {} {
    global cfgContents env
    if {![info exists cfgContents(pgp)]} {
	global cfgFileName
	nsbderror  \
"\"pgp\" configuration keyword not set.  A complete path to PGP is required
  in the \"pgp:\" keyword in $cfgFileName for security. 
  If you do not have PGP, you can use the \"-ignoreSecurity\" command line
  option when installing packages, but then you throw out security.  To create
  an unsigned '.nsb' file you can use the \"-unsigned\" command line option."
    }
    if {[info exists cfgContents(pgppath)]} {
	set env(PGPPATH) $cfgContents(pgppath)
    }
    return "$cfgContents(pgp)"
}

#
# Get the name of the term program
#
proc getterm {} {
    global cfgContents
    if {[info exists cfgContents(termCommand)]} {
	return $cfgContents(termCommand)
    }
    return "xterm"
}

#
# Get the name of the shell
#
proc getshell {} {
    global env
    if {[info exists env(SHELL)]} {
	return $env(SHELL)
    }
    return "sh"
}

#
# Apply pgp signature to $inFile, leaving signed file in $outFile
#
proc pgpSignFile {inFile outFile} {
    set ttyfd stdout
    set pgp [getpgp]
    # don't use progressmsg here, want to prompt the user
    puts $ttyfd "Invoking pgp to create signature for $outFile"
    set inFile [expandTildes $inFile]
    set outFile [expandTildes $outFile]
    # pgp options: s - sign; a - ascii armor; t - make independent
    #     of platform end-of-line
    set fd [nsbdcatch {open "|$pgp +verbose=0 -sat $inFile" "r"}]
    alwaysEvalFor $outFile {catch {close $fd}} {
	set sigfile ""
	set getpass 0
	# eat up first line if it is blank
	gets $fd line
	if {$line != ""} {
	    puts $ttyfd $line
	}
	while {[gets $fd line] >= 0} {
	    # The prompts may be different depending on the version of PGP.
	    # Tested with pgp 2.6.2 and viacrypt pgp 4.0 business edition.
	    #
	    set firstfour [lrange $line 0 3]
	    if {$firstfour == "Key for user ID"} {
		# pgp 2.6.2
		puts $ttyfd $line
		set getpass 1
		continue
	    }
	    if {$getpass && ($line != "")} {
		# this line should have been blank; in case not really blank
		puts $ttyfd $line
	    }
	    set firsttwo [lrange $line 0 1]
	    if {$getpass || ($firsttwo == "with userid:")} {
		# first case pgp 2.6.2, second case pgp 4.0
		puts -nonewline $ttyfd "Enter Passphrase: "
		flush stdout
		set getpass 0
		continue
	    }
	    set firstone [lindex $line 0]
	    if {$firstone == "\007Error:"} {
		# pgp 4.0
		puts $ttyfd $line
		set getpass 1
		continue
	    }
	    set firstthree [lrange $line 0 2]
	    if {($firstthree == "Enter pass phrase:") ||
		    ($firsttwo == "Enter Passphrase:") ||
			($firsttwo == "Enter Passphrase:\007")} {
		# first form pgp 2.6.2, second and third forms pgp 4.0
		# eat up the prompt that PGP put out, but put out a
		#  newline because echoing was turned off for password
		puts $ttyfd ""
		continue
	    }
	    if {($firstthree == "To sign file:") ||
		    ($firstthree == "Error code =")} {
		# pgp 4.0
		continue
	    }
	    if {$firstthree == "Clear signature file:"} {
		set sigfile [lindex $line 3]
		addToScratchFiles $sigfile
		progressmsg "Successfully created $outFile"
		continue
	    }
	    if {$line != ""} {
		puts $ttyfd $line
	    }
	}
    }
    if {$sigfile == ""} {
	nsbderror "error creating pgp signature"
    }
    # Need to make a copy of the file simply because pgp always changes
    #  the umask to be 077, and want default umask.  Otherwise, could
    #  use "-o" option on pgp (which still appends ".asc", so would
    #  still need to at least rename the file and would need to be careful
    #  to put it in the same filesystem as the destination or rename will
    #  do a copy too.)
    withOpen fdout $outFile "w" {
	withOpen fdin $sigfile "r" {
	    copychannel $fdin $fdout
	}
    }
    scratchClean $sigfile
}

#
# Check the pgp signature in $sigFile.  sigFile may be a temporary name, so
#   use forFile as the name for error messages instead; if forFile is empty
#   no name will be added to the error messages.
# Return value is in 4 parts:
#   1. A boolean (1 or 0) that indicates if the signature was good or not.
#   2. The user ID of the signature, if it could be determined (good or bad).
#      If the public key could not be found, the user ID is "*UNKNOWN*".
#   3. The date/time the signature was made, if it could be determined.
#   4. Any warning or error messages, possibly including newlines.

proc pgpCheckFile {signedFile {forFile ""}} {
    # Note: there are two reasons to separate the plaintext from the signature
    #  into two temporary files.  One is to allow multiple signatures in the
    #  future, and the other is that with a combined file pgp will attempt to
    #  create a plaintext file without the signature.  The only way around that
    #  is to apply the option "-o $signedFile" so the default filename will be
    #  the same as $inFile and it won't overwrite an existing file and it will
    #  return an error code.
    progressmsg "Checking PGP signature on [expr {$forFile == "" ? $signedFile : $forFile}]"
    set pgp [getpgp]
    set signedFile [expandTildes $signedFile]
    set sigFile [scratchAddName "sig"]
    set plainFile [scratchAddName "pln"]
    set foundsig 0
    withOpen fd $signedFile "r" {
	withOpen fdplain $plainFile "w" {
	    gets $fd line
	    while {[string range $line 0 4] == "-----"} {
		# skip beginning lines that start with ----
		gets $fd line
	    }
	    while {![eof $fd] && ([string trim $line] == "")} {
		# skip beginning blank lines
		gets $fd line
	    }
	    set blanks ""
	    while {![eof $fd]} {
		if {[string trim $line] == ""} {
		    append blanks "$line\n"
		    gets $fd line
		    continue
		}
		if {[string range $line 0 4] == "-----"} {
		    # now into ending signature blocks
		    break
		}
		if {$blanks != ""} {
		    # save blank lines in the middle of the file
		    puts -nonewline $fdplain $blanks
		    set blanks ""
		}
		puts $fdplain $line
		gets $fd line
	    }
	    # skip ending blank lines
	    while {![eof $fd]} {
		if {[string range $line 0 23] == "-----BEGIN PGP SIGNATURE"} {
		    withOpen fdsig $sigFile "w" {
			while {![eof $fd]} {
			    puts $fdsig $line
			    gets $fd line
			    if {[string range $line 0 21] == "-----END PGP SIGNATURE"} {
				set foundsig 1
				puts $fdsig $line
				break
			    }
			}
		    }
		}
		gets $fd line
	    }
	}
    }
    if {!$foundsig} {
	scratchClean "pln"
	scratchClean "sig"
	return {0 "" "" "Did not find complete signature block"}
    }

    set fd [nsbdcatch {open "|$pgp +verbose=0 +batchmode $sigFile $plainFile" "r"}]
    set good 0
    set user ""
    set date ""
    set warning ""
    alwaysEvalFor "" {catch {close $fd}} {
	while {[gets $fd line] >= 0} {
	    # The prompts may be different depending on the version of PGP.
	    # Tested with pgp 2.6.2 and viacrypt pgp 4.0 business edition.
	    switch -regexp -- $line {
		"^Good signature from user " {
		    # pgp 2.6.2
		    set good 1
		    regexp {^Good signature from user "([^"]*)"} $line x user
		}
		"^Good signature from user:  " {
		    # pgp 4.0
		    set good 1
		    regexp {^Good signature from user:  ([^"]*)} $line x user
		}
		"^Bad signature from user " {
		    # pgp 2.6.2
		    set good 0
		    regexp {^Bad signature from user "([^"]*)"} $line x user
		}
		"Bad signature from user:  " {
		    # pgp 4.0
		    set good 0
		    regexp {Bad signature from user:  ([^"]*)} $line x user
		}
		"Signature made  " {
		    # pgp 4.0
		    regexp {Signature made  ([^G]* GMT)} $line x date
		    set date [canonicalizeTime $date]
		}
		"^Signature made " {
		    # pgp 2.6.2 
		    regexp {^Signature made ([^G]* GMT)} $line x date
		    set date [canonicalizeTime $date]
		}
		"^\007" -
		"^WARNING:" {
		    # seen in both pgp 2.6.2 and 4.0
		    if {$line == "\007"} {
			gets $fd msg
		    } elseif {[string index $line 0] == "\007"} {
			set msg [string range $line 1 end]
		    } else {
			set msg $line
		    }
		    if {[lindex $msg 0] == "ERROR:"} {
			lappend warning [lrange $msg 1 end]
		    } else {
			if {[lindex $msg 0] == "WARNING:"} {
			    set msg [lrange $msg 1 end]
			}
			if {![regexp "Output file.*already exists" $msg] &&
			    ![regexp "Key matching expected.*not found" $msg] &&
			    ![regexp "Problem with clear result" $msg]} {
			    # skip expected messages in pgp 2.6.2 and pgp 4.0
			    lappend warning $msg
			}
			while {([gets $fd msg] > 0) && 
						![regexp {^[A-Z]} $msg]} {
			    # pgp 2.6.2 follows warning by blank lines, but the
			    #  pgp 4.0 warnings that I've seen are followed by
			    #  lines that begin with upper case letters, and
			    #  they're lines that I don't care about
			    lappend warning $msg
			}
		    }
		}
	    }
	}
    }
    set warning [join $warning "\n"]
    if {!$good && ([lrange $warning 0 4] == "Can't find the right public")} {
	set user "*UNKNOWN*"
    }
    scratchClean "sig"
    scratchClean "pln"
    return [list $good $user $date $warning]
}

#
# Return 1 if id is known to PGP, otherwise return 0
#

proc pgpIDKnown {id} {
    set pgp [getpgp]
    set fd [nsbdcatch {open "|$pgp +verbose=0 +batchmode -kv \"$id\"" "r"}]
    alwaysEvalFor "" {catch {close $fd}} {
	while {[gets $fd line] >= 0} {
	    if {($line == "1 matching key found.") ||
		($line == "matching keys found:  1")} {
		    return 1
	    }
	}
    }
    return 0
}

#
# Pop up a new window to interactively run pgp to add a key
#
proc pgpAddKey {id keyfile} {
    set pgp [getpgp]
    set scratch1 [scratchAddName "a.pgp"]
    # all tested pgp's create an empty a.bak file
    scratchAddName "a.bak"
    # pgp 4.0 can't handle a scratch2 name with a ".pgp" extension,
    #   but pgp 2.6.2 can
    set scratch2 [scratchAddName "key2"]
    # pgp 2.6.2 adds a ".pgp" extension to key2, but pgp 4.0 does not
    scratchAddName "key2.pgp"
    eval exec "[getterm] -T {PGP for NSBD} -e [getshell] -c {
	echo Calling PGP to extract key
	echo \"    \" \"$id\"
	# need to temporarily add to $scratch1 because pgp 2.6.2 cannot
	#  extract an id from an ascii-armored keyfile; pgp 4.0 can
	$pgp +verbose=0 +batchmode -ka $keyfile $scratch1 >/dev/null
	# first $scratch2 is for pgp 2.6.2 and -o $scratch2 is for pgp 4.0
	$pgp +verbose=0 +batchmode -kx \"$id\" $scratch2 $scratch1 -o $scratch2
	echo Calling PGP to add the key
	# pgp 4.0 doesn't handle putting -kvc keyring at the end of command
	#  line although pgp 2.6.2 does, so using +pubring instead
	$pgp +verbose=0 +pubring=$scratch2 -kvc \"$id\"
	$pgp +verbose=0 -ka $scratch2
	echo Hit return to continue...
	read DUMMY
    }"
    scratchClean "a.pgp"
    scratchClean "a.bak"
    scratchClean "key2"
    scratchClean "key2.pgp"
}
