#
# NSBD interface to tcl's standard (7.7 or greater) http library
# File created by Dave Dykstra <dwd@bell-labs.com>, 09 Jan 1997
#
#
# Copyright (C) 1996-7 by Dave Dykstra and Lucent Technologies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#  
# If those terms are not sufficient for you, contact the author to 
# discuss the possibility of an alternate license.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

proc nsbdHttpProxyFilter {host} {
    # Warning: includes hacks to allow support for ftp through an http
    #  proxy; has knowledge of the implementation of the http library.
    #  Specifically, it knows that the variable name in http_get that
    #  holds the url is "url", and it knows that http_get directly calls
    #  this filter with no intermediate routines such that the variables
    #  from urlGet are exactly 2 levels up
    upvar 2 reallyFtp reallyFtp
    upvar 2 proxyHostAndPort proxyHostAndPort
    if {[info exists reallyFtp]} {
	upvar url url
	regsub "^http:" $url "ftp:" url
    }
    if {$proxyHostAndPort != ""} {
	debugmsg "  using proxy $proxyHostAndPort"
    }
    return $proxyHostAndPort
}

http_config -proxyfilter nsbdHttpProxyFilter

#
# Interface to http_get that also allows ftp through an http proxy
#
proc urlGet {args} {
    set url [HMmap_reply [lindex $args 0]]
    if {[lsearch -exact $args "-query"] >= 0} {
	progressmsg "POSTing to $url"
    } else {
	progressmsg "Fetching $url"
    }
    global cfgContents
    if {[regexp -nocase "^http://(\[^/:\]+)" $url x host]} {
	if {![info exists cfgContents(http_proxy)]} {
	    set proxyHostAndPort ""
	} elseif {[info exists cfgContents(no_proxy)]} {
	    # look through list of domains for which to not use the proxy
	    foreach domain $cfgContents(no_proxy) {
		if {[regexp -nocase "\.$domain\$" $host] || \
				[regexp -nocase "^$domain\$" $host]} {
		    set proxyHostAndPort ""
		    break
		}
	    }
	}
	if {![info exists proxyHostAndPort]} {
	    if {![regexp -nocase "^http://(\[^:\]*)(:(.*))?" \
			$cfgContents(http_proxy) x phost y pport]} {
		nsbderror {http_proxy must be in form http://proxyhost[:portno] (default portno is 8080)}
	    }
	    if {$pport == ""} {
		set pport 8080
	    }
	    set proxyHostAndPort "$phost $pport"
	}
    } elseif {[regsub "^ftp://" $url "http://" url]} {
	set reallyFtp 1
	if {![info exists cfgContents(ftp_proxy)]} {
	    nsbderror "ftp only supported through a proxy, and ftp_proxy not set"
	}
	if {![regexp -nocase "^http://(\[^:\]*)(:(.*))?" \
			$cfgContents(ftp_proxy) x phost y pport]} {
	    nsbderror {ftp_proxy must be in form http://proxyhost[:portno] (default portno is 8080)}
	}
	if {$pport == ""} {
	    set pport 8080
	}
	set proxyHostAndPort "$phost $pport"
    } else {
	nsbderror "url must begin with http:// or ftp://: $url"
    }
    # set up for possible error opening socket
    if {$proxyHostAndPort != ""} {
	set for "proxy [join $proxyHostAndPort ":"]"
    } else {
	set for ""
        regexp -nocase "^http://(\[^/:\]+(:\[0-9\]+)?)" $url x for
    }
    return [alwaysEvalFor "$for" {} {
		nsbdcatch {eval http_get $url [lrange $args 1 end]}
	    }]
}

#
# Check results of an http operation for errors.   If there were any,
#  raise a TCL error condition
#
proc httpCheck {token url} {
    upvar #0 $token state
    if {![info exists state(http)]} {
	# I don't know if there are any other reasons this could be,
	#  but this is the only one I've seen, and there's no other
	#  way to tell a timeout I don't think
	nsbderror "timed out getting $url header"
    }
    set httpCode [lindex $state(http) 1]
    if {$httpCode >= 400} {
	nsbderror "error getting $url: $state(http)"
    }
    if {($state(totalsize) > 0) && ($state(currentsize) < $state(totalsize))} {
	nsbderror "$url length incorrect: expected $state(totalsize), got $state(currentsize)"
    }
}

#
# Copy a url to a local file.  Let the end-of-line translation be
#   automatically selected (text file is assumed).
# If the "if-modified-since" is set (to a canonical time), will ask the
#   server to only retrieve the file if it has been modified since that
#   time; if it had not, will return an empty string, otherwise will return
#   the server time in PGP format if available, or the current local time.
# If not successful, an error will be raised.
#

proc urlCopy {url localfile {ifModifiedSince ""} {extraArgs ""}} {
    if {$ifModifiedSince != ""} {
	lappend extraArgs -headers [list "If-Modified-Since" $ifModifiedSince]
    }
    if {$localfile != "-"} {
	set fd [nsbdcatch {open $localfile "w"}]
	alwaysEvalFor "" {close $fd} {
	    set token [eval "urlGet $url -channel $fd" $extraArgs]
	    alwaysEvalFor $localfile {} {
		http_wait $token
	    }
	}
    } else {
	set fd stdout
	set token [eval "urlGet $url -channel $fd" $extraArgs]
	http_wait $token
    }
    httpCheck $token $url
    upvar #0 $token state
    if {$ifModifiedSince != ""} {
	set httpCode [lindex $state(http) 1]
	if {$httpCode == 304} {
	    progressmsg "  wasn't modified since $ifModifiedSince, skipping"
	    http_reset $token
	    return ""
	}
    }
    array set meta $state(meta)
    http_reset $token
    if {[info exists meta(Date)]} {
	return [canonicalizeTime $meta(Date)]
    } else {
	# use our own time if can't find it from the server
	return [currentTime]
    }
}

#
# If urlOrFile is a supported URL, copy it to a scratch local file and return
#  the name of the local file.  Otherwise, return the filename itself.
# Also return the server time of the URL file as a second return value
#  if known.
# 

proc urlFile {urlOrFile {ifModifiedSince ""}} {
    if {[regexp {://} $urlOrFile]} {
	set scratchName [scratchAddName "ucp"]
	return [list $scratchName \
		[urlCopy $urlOrFile $scratchName $ifModifiedSince]]
    }
    return [list $urlOrFile]
}

#
# Copy a url to a local file while calculating the message digest checksum.  
# mdType is message digest type
# If successful, return a list of these items:
#   1. Length of the file in bytes
#   2. Message digest checksum in hexadecimal ascii
#   3. File open permissions mode in octal
# If not successful, an error will be raised.
#

proc urlMdCopy {url mdType localfile {mode ""}} {
    if {$mode == ""} {
	set mode "0666"
    }
    set fd [nsbdcatch {open $localfile "w" $mode}]
    set mdDescriptor [$mdType -init]
    alwaysEvalFor "" {close $fd} {
	set token [urlGet $url -handler "urlMdCopyHandler $fd"]
	upvar #0 $token state
	set state(mdType) $mdType
	set state(mdDescriptor) $mdDescriptor
	alwaysEvalFor $localfile {} {
	    http_wait $token
	}
    }
    httpCheck $token $url
    http_reset $token
    return [$mdType -final $mdDescriptor]
}

#
# This is called whenever data is available on the url.  Return the
#  number of bytes read.
#
proc urlMdCopyHandler {fd socket token} {
    upvar #0 $token state
    if {![info exists state(afterFirstBlock)]} {
	# always want binary, no matter what the Content-type is
	fconfigure $socket -translation binary
	fconfigure $fd -translation binary
	set state(afterFirstBlock) 1
    }
    if {[info exists state(-blocksize)]} {
	set blocksize $state(-blocksize)
    } else {
	set blocksize 8192
    }
    return [$state(mdType) -update $state(mdDescriptor) -chan $socket \
			-copychan $fd -maxbytes $blocksize]
}

#
# Fetch multiple files in one http connection.
# Assumes talking to a CGI script on the other side that accepts file
#  names in a POST to its standard input, and then for each file puts out
#  the filename in a Content-Name: header and the length in Content-Length:
#  followed by a blank line followed by the data for the file.
# mdType is message digest type
# pathsInfo is a list of info about each path.  Info is a list of
#   1. fromPath - path to retrieve
#   2. toPath - path to store into
#   3. mode - mode of file to store
#   4. mdData - the expected message digest data for the file (length and
#		message digest checksum)
#
proc urlMultiMdCopy {url mdType pathsInfo} {
    set numpaths [llength $pathsInfo]
    if {$numpaths == 0} {
	return ""
    }
    set paths ""
    set sep ""
    foreach pathInfo $pathsInfo { 
	append paths $sep[lindex $pathInfo 0]
	set sep "\n"
    }
    set token [urlGet $url -query $paths -handler urlMultiMdCopyHandler]
    upvar #0 ${token}_multi multi
    set multi(mdType) $mdType
    alwaysEvalFor "" {
			if {[info exists multi(fd)]} {close $multi(fd)}
			set multistate $multi(state)
			set pathnum $multi(pathnum)
			set remaining $multi(remaining)
			catch {unset multi}
		     } {
	set multi(state) init
	set multi(pathnum) 0
	set multi(remaining) 0
	set multi(pathsInfo) $pathsInfo
	http_wait $token
    }
    httpCheck $token $url
    http_reset $token
    if {$multistate != "want-Keyword"} {
	if {($multistate == "reading-Content") && ($remaining > 0)} {
	    nsbderror "multi-fetch had premature end of file, $remaining bytes remaining in file $pathnum"
	}
	nsbderror "data was not in multiget format (in state $multistate)"
    }
    if {$pathnum != $numpaths} {
	nsbderror "multi-fetched $pathnum files but expected $numpaths"
    }
}

#
# This is called whenever data is available on the url.  Return the
#  number of bytes read.
#
proc urlMultiMdCopyHandler {socket token} {
    upvar #0 $token state
    upvar #0 ${token}_multi multi
    set mdType $multi(mdType)
    if {$multi(state) == "init"} {
	fconfigure $socket -translation binary
	set multi(state) want-Keyword
    }
    if {$multi(state) != "reading-Content"} {
	# note that the line may have a carriage return in it
	# can't let tcl strip that out because it would mess up the byte count
	set bytes [gets $socket line]
	set line [string trimright $line "\r"]
	set keyword ""
	if {[regexp {^([^:]+):(.+)$} $line x keyword value]} {
	    set value [string trim $value]
	} elseif {$line != ""} {
	    nsbderror "bad multiget header format in state $multi(state), got\n    [string range $line 0 60]"
	}
	if {[regexp -nocase {^x-multiget-error$} $keyword]} {
	    nsbderror "error from multiget remote:\n    $value"
	}
	if {$multi(state) == "want-Keyword"} {
	    if {$line != ""} {
		set multi(state) want-Content-Name
	    }
	}
	if {$multi(state) == "want-Content-Name"} {
	    set pathInfo [lindex $multi(pathsInfo) $multi(pathnum)]
	    if {$pathInfo == ""} {
		set pathInfo [list ""]
	    }
	    foreach {fromPath toPath mode mdData} $pathInfo {}
	    if {$line == ""} {
		nsbderror "Content-Name header not found (expected $fromPath)"
	    }
	    if {[regexp -nocase {^content-name$} $keyword]} {
		if {$fromPath != $value} {
		    nsbderror "url multi-fetch retrieved unexpected path $value, expected $fromPath"
		}
		progressmsg "Multi-fetching $fromPath"
		set multi(fromPath) $fromPath
		set multi(toPath) $toPath
		set multi(mdData) $mdData
		if {$mode == ""} {
		    set mode 0666
		}
		set multi(mdDescriptor) [$mdType -init]
		set multi(fd) [nsbdcatch {open $toPath "w" $mode}]
		fconfigure $multi(fd) -translation binary
		set multi(state) want-Content-Length
		incr multi(pathnum)
	    }
	} elseif {$multi(state) == "want-Content-Length"} {
	    if {$line == ""} {
		nsbderror "Content-Length header for $multi(fromPath) not found"
	    }
	    if {[regexp -nocase {^content-length$} $keyword]} {
		if {![regexp {^[0-9]+$} $value]} {
		    nsbderror "Content-Length value \"$value\" is not a number"
		}
		set multi(remaining) $value
		set multi(state) "want-blank-line"
	    }
	} elseif {$multi(state) == "want-blank-line"} {
	    if {$line == ""} {
		set multi(state) reading-Content
	    }
	}
	return $bytes
    }

    if {[info exists state(-blocksize)]} {
	set blocksize $state(-blocksize)
    } else {
	set blocksize 8192
    }
    if {$blocksize > $multi(remaining)} {
	set blocksize $multi(remaining)
    }
    set bytes [$mdType -update $multi(mdDescriptor) -chan $socket \
			-copychan $multi(fd) -maxbytes $blocksize]
    if {[incr multi(remaining) -$bytes] == 0} {
	close $multi(fd)
	unset multi(fd)
	if {$multi(mdData) != ""} {
	    set mdData [$mdType -final $multi(mdDescriptor)]
	    compareMdDataFor $multi(fromPath) $multi(mdData) $mdData
	}
	set multi(state) want-Keyword
    }
    return $bytes
}

#
# Compare expected message digest data to received data for $path 
# Raise an error if they don't match
#
proc compareMdDataFor {path expectedMdData mdData} {
    set expectedLength [lindex $expectedMdData 0]
    set expectedMd [lindex $expectedMdData 1]
    set length [lindex $mdData 0]
    set md [lindex $mdData 1]
    if {$length != $expectedLength} {
	nsbderror "wrong length of $path: got $length, expected $expectedLength"
    }
    if {[string compare $md $expectedMd] != 0} {
	nsbderror "wrong message digest for $path: got $md, expected $expectedMd"
    }
}

# From "Uhler's html_library", as posted by Steve Ball on comp.lang.tcl

# do x-www-urlencoded character mapping
# The spec says: "non-alphanumeric characters are replaced by '%HH'"
 
set HMalphanumeric      a-zA-Z0-9       ;# definition of alphanumeric character class
# add other characters I don't want translated
set HMalphanumeric	-${HMalphanumeric}*,./:?_~=
for {set i 1} {$i <= 256} {incr i} {
    set c [format %c $i]
    if {![string match \[$HMalphanumeric\] $c]} {
        set HMform_map($c) %[format %.2x $i]
    }
}
 
# These are handled specially
array set HMform_map {
    " " +   \n %0d%0a
}
 
# 1 leave alphanumerics characters alone
# 2 Convert every other character to an array lookup
# 3 Escape constructs that are "special" to the tcl parser
# 4 "subst" the result, doing all the array substitutions
 
proc HMmap_reply {string} {
    global HMform_map HMalphanumeric
    regsub -all \[^$HMalphanumeric\] $string {$HMform_map(&)} string
    regsub -all \n $string {\\n} string
    regsub -all \t $string {\\t} string
    regsub -all {[][{})\\]\)} $string {\\&} string
    return [subst $string]
}
