#
# General TCL utilities for use in NSBD
# File created by Dave Dykstra <dwd@bell-labs.com>, 12 Nov 1996
#
#
# Copyright (C) 1996-7 by Dave Dykstra and Lucent Technologies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#  
# If those terms are not sufficient for you, contact the author to 
# discuss the possibility of an alternate license.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

#
# alwaysEvalFor: always evaluate the alwaysscript after the
#    evalscript even if there was an error in the evalscript.
#    errorPrefix, if nonempty, is a prefix to put on any error message.
#
proc alwaysEvalFor {errorPrefix alwaysscript evalscript} {
    set code [catch {uplevel $evalscript} string]
    uplevel $alwaysscript
    if {($code == 1) && ($errorPrefix != "")} {
	set string [addErrorPrefix $errorPrefix $string]
    }
    global errorInfo errorCode
    return -code $code -errorinfo $errorInfo -errorcode $errorCode $string
}

#
# add an error prefix, trying to avoid making lines that are too long
#
proc addErrorPrefix {prefix string} {
    set strlen [string first $string "\n"]
    if {$strlen < 0} {
	set strlen [string length $string]
    }
    if {([string length $prefix] + $strlen) > 72} {
	return "$prefix:\n  $string"
    }
    return "$prefix: $string"
}

#
# Convenient interface to alwaysEvalFor for most common usage: opening a file.
# fdName is the name of a variable to use for the open file descriptor,
#   filename is the name of the file, access is the access string ("r", "w",
#   etc.", and $evalscript is the script to execute while the file is open.
#
proc withOpen {fdName filename access evalscript} {
    set fd [uplevel [list open $filename $access]]
    uplevel "set $fdName $fd"
    set code [catch {uplevel $evalscript} string]
    uplevel "close \$$fdName"
    if {$code == 1} {
	set string [addErrorPrefix $filename $string]
    }
    global errorInfo errorCode
    return -code $code -errorinfo $errorInfo -errorcode $errorCode $string
}

#
# Append items in arg1 and args to list if and only if the items are not
#   already on the list.
# Return the number of items that were not already on the list
#
proc luniqueAppend {varName arg1 args} {
    upvar $varName var
    set changes 0 
    if {![info exists var] || ([lsearch -exact $var $arg1] < 0)} {
	set lastval [lappend var $arg1]
	incr changes
    }
    foreach arg $args {
	if {[lsearch -exact $var $arg] < 0} {
	    set lastval [lappend var $arg]
	    incr changes
	}
    }
    return $changes
}

#
# Substitute all occurrences of the regular expression $exp in all
#   the strings in $strings with all of the substitute specifications
#   in subSpecs.  Return a list of all the substitute strings.
#
proc multiSubstitute {exp strings subSpecs} {
    set newStrings ""
    foreach string $strings {
	if {[regexp $exp $string]} {
	    foreach subSpec $subSpecs {
		regsub -all $exp $string $subSpec newString
		lappend newStrings $newString
	    }
	} else {
	    lappend newStrings $string
	}
    }
    return $newStrings
}

#
# Return 1 if file1 and file2 have the same contents, otherwise 0.
# The two files must be text files, not binary files.
#
proc compareFiles {file1 file2} {
    nsbdcatch {file stat $file1 stat1}
    nsbdcatch {file stat $file2 stat2}
    if {$stat1(size) != $stat2(size)} {
	return 0
    }
    withOpen fd1 $file1 "r" {
	withOpen fd2 $file2 "r" {
	    while {[set buf [read $fd1 1024]] != ""} {
		if {[read $fd2 1024] != $buf} {
		    return 0
		}
	    }
	}
    }
    return 1
}

#
# Return complete path to executable in PATH, or empty if not found
#
proc whereExecutable {executable} {
    global env
    foreach dir [split $env(PATH) ":"] {
	if {$dir == ""} {
	    set dir "."
	}
	set f $dir/$executable
	if {[file executable $f] && [file isfile $f]} {
	    return $f
	}
    }
    return ""
}

#
# Expand tildes in path name.  I can't find an easy way to do this
#  directly in tcl, so need to cd to the directory and and get pwd.
#

proc expandTildes {filename} {
    if {[string index $filename 0] != "~"} {
	return $filename
    }
    set savepwd [pwd]
    # create the directory if it doesn't yet exist
    set dirname [file dirname $filename]
    nsbdcatch {file mkdir $dirname}
    cd $dirname
    set filename [file join [pwd] [file tail $filename]]
    cd $savepwd
    return $filename
}

#
# figure out the verboseLevel
#
proc verboseLevel {} {
    global cfgContents
    if {[info exists cfgContents(verbose)]} {
	return $cfgContents(verbose)
    } else {
	return 6
    }
}

#
# Return a filename with a prefix on the tail
#
proc addFilePrefix {fileName prefix} {
    set dirName [file dirname $fileName]
    if {$dirName == "."} {
	set dirName ""
    }
    return [file join $dirName $prefix[file tail $fileName]]
}

#
# log a message, which may contain newlines.  If $level is less than or
#   equal to logLevel configuration keyword, write it to the logFile with
#   a timestamp.  If the message is less than or equal to the verbose level,
#   also write interactively.
# levels:
#	0 - error
#	1 - warning
#	2 - notice from maintainer
#	3 - update
#	5 - progress info
#	7 - debug
#
proc logmsg {level msg} {
    set logged 0
    global cfgContents
    if {[info exists cfgContents(logFile)]} {
	set logFile $cfgContents(logFile)
    } else {
	set logFile "updates.log"
    }
    if {[info exists cfgContents(logLevel)]} {
	set logLevel $cfgContents(logLevel)
    } else {
	set logLevel 4
    }
    if {$level <= $logLevel} {
	set fname [file join $cfgContents(nsbdpath) $logFile]
	trimOldLog $fname
	withOpen fd $fname "a" {
	    set now [logTime]
	    foreach line [split $msg "\n"] {
		puts $fd "$now $line"
	    }
	}
    }
    global messagesWindow
    if {($messagesWindow != "") && 
		(($level <= 2) || ($level <= [verboseLevel]))} {
	$messagesWindow insert end "$msg\n"
	$messagesWindow see end
	update idletasks
    } elseif {$level <= [verboseLevel]} {
	if {$level == 0} {
	    puts -nonewline stderr "nsbd: "
	}
	puts stderr $msg
	flush stderr
    }
}

#
# Log an error message.
#
proc errormsg {msg} {
    global errorInfo errorCode
    set savErrorInfo $errorInfo
    set savErrorCode $errorCode
    set code [catch {logmsg 0 $msg} string]
    if {$code > 0} {
	puts stderr "Error in logging error message:"
	puts $errorInfo
	puts stderr "Original error message was: $msg"
    }
    set errorInfo $savErrorInfo
    set errorCode $savErrorCode
}

#
# Log a warning message.
#
proc warnmsg {msg} {
    logmsg 1 $msg
    global messagesWindow
    if {$messagesWindow != ""} {
	set ans [tk_dialog .warnmsg "Warning" $msg warning 0 "Ok" "Exit"]
	if {$ans == 1} {
	    nsbdExit 1
	}
    }
}

#
# Log other kinds of messages
#
proc noticemsg {msg} {
    logmsg 2 $msg
}

proc updatemsg {msg} {
    logmsg 3 $msg
}

proc progressmsg {msg} {
    logmsg 5 $msg
}

proc debugmsg {msg} {
    logmsg 7 $msg
}

#
# Check for an old log file the first time the log file is accessed.
# If the first entry is more then logDays old, rename existing log with an
#   "old" prefix.

proc trimOldLog {fname} {
    global logFileHasBeenChecked
    if {[info exists logFileHasBeenChecked]} {
	return
    }
    set logFileHasBeenChecked 1

    global cfgContents
    if {[info exists cfgContents(logDays)]} {
	set logDays $cfgContents(logDays)
    } else {
	set logDays 7
    }
    if {$logDays == 0} {
	return
    }
    catch {file stat $fname statb}
    if {![info exists statb(mtime)]} {
	return
    }
    set modtime $statb(mtime)
    unset statb
    withOpen fd $fname "r" {
	gets $fd line
    }
    set firstTime [lrange $line 0 1]
    if {[catch {scanAnyTime $firstTime} firstSeconds] != 0} {
	set firstSeconds 0
    }
    set nowSeconds [clock seconds]
    set daysOld [expr {($nowSeconds - $firstSeconds) / (24 * 60 * 60)}]
    if {$daysOld <= $logDays} {
	debugmsg "$fname was $daysOld days old, not saving"
	return
    }
    set oldfname [addFilePrefix $fname "old"]
    lockFile $fname
    catch {file stat $fname statb}
    if {![info exists statb(mtime)] || ($modtime != $statb(mtime))} {
	unlockFile $fname
	debugmsg "Didn't lock $fname before it was modified, skipping this time"
	return
    }
    file delete -force $oldfname
    nsbdcatch {file rename $fname $oldfname}
    unlockFile $fname
    if {$firstSeconds == 0} {
	progressmsg "Couldn't scan time on first line of $fname\n   renamed it to $oldfname"
    } else {
	progressmsg "Log file $fname was $daysOld days old\n  renamed it to $oldfname"
    }
}

#
# Lock a file.  Create lock file with exclusive create permission.  If
#  file already exists, try up to 5 times one second apart.  If the
#  modification time of the lock file or $file changes, assume another
#  program is making progress and reset the count.  If neither changes
#  after 5 seconds, assume the other program has died and forcibly remove
#  the lock before retrying.
# NOTE: perhaps this should be changed to use the Unix 'fcntl' F_SETLK
#  facility.  I'm not sure how portable that would be though.  It's nice
#  in that it can distinguish between read locks and read/write locks.
#

proc lockFile {filename} {
    debugmsg "Obtaining lock on $filename"
    set lockname [addFilePrefix $filename LCK]
    set tries 0
    set modtime 0
    while {$tries <= 5} {
	set code [catch {set fd [open $lockname "WRONLY CREAT EXCL"]} why]
	if {$code == 0} {
	    puts $fd [pid]
	    close $fd
	    addToScratchFiles $lockname
	    return
	}
	incr tries
	after 1000
	foreach file {$lockname $filename "last"} {
	    if {$file == "last"} {
		if {$tries >= 5} {
		    progressmsg "Overriding $lockname, lock held at least 5 seconds"
		    nsbdcatch {file delete $lockname}
		    set tries 0
		}
	    } else {
		catch {unset statb}
		catch {file stat $file statb}
		if {[info exists statb(mtime)]} {
		    if {$statb(mtime) > $modtime} {
			set modtime $statb(mtime)
			set tries 0
			break
		    }
		}
	    }
	}
    }
    nsbderror "$why"
}

#
# Remove the lock file
#
proc unlockFile {filename} {
    debugmsg "Releasing lock on $filename"
    scratchClean [addFilePrefix $filename LCK]
}

#
# Center the window on the screen.  This code is borrowed from tk_dialog
#
proc tkCenter {{win "."}} {
    wm withdraw $win
    update idletasks
    set x [expr {([winfo screenwidth $win] - [winfo reqwidth $win]) / 2}]
    set y [expr {([winfo screenheight $win] - [winfo reqheight $win]) / 2}]
    wm geom $win +$x+$y
    wm deiconify $win
}

#
# pop up messages window
#
proc makeMessagesWindow {} {
    global messagesWindow

    set mw ".messages"
    toplevel $mw
    wm title $mw "NSBD messages"
    text $mw.text -width 80 -height 35 -yscrollcommand "$mw.scroll set"
    scrollbar $mw.scroll -command "$mw.text yview" -takefocus 0
    pack $mw.scroll -side right -fill y
    pack $mw.text -side left -fill x -expand 1
    tkCenter $mw
    set messagesWindow $mw.text
    update idletasks
}

#
# remove messages window
#
proc deleteMessagesWindow {} {
    global messagesWindow

    destroy $messagesWindow
    set messagesWindow ""
}

#
# raise the messages window
#
proc raiseMessagesWindow {} {
    global messagesWindow

    wm withdraw .messages
    wm deiconify .messages
    update idletasks
}

#
# Compare two version numbers.  Return negative if version1 is less than
#   version2, 0 if the versions are equal, or positive if version1 is greater
#   than version2.
#
proc compareVersions {version1 version2} {
    while {[string compare $version1 $version2] != 0} {
	set n1 ""; set a1 ""; set r1 ""
	set n2 ""; set a2 ""; set r2 ""
	regexp {^([0-9]*)([^0-9]*)(.*)} $version1 x n1 a1 r1
	regexp {^([0-9]*)([^0-9]*)(.*)} $version2 x n2 a2 r2
	if {$n1 < $n2} {
	    return -1
	}
	if {$n1 > $n2} {
	    return 1
	}
	# alpha or beta suffixes are considered to be less than
	#   a number without an alpha or beta suffix
	set pre1 [expr {[string first [string index $a1 0] "AaBb"] >= 0}]
	set pre2 [expr {[string first [string index $a2 0] "AaBb"] >= 0}]
	if {$pre1 && !$pre2} {
	    return -1
	}
	if {!$pre1 && $pre2} {
	    return 1
	}
	if {[set comp [string compare $a1 $a2]] != 0} {
	    return $comp
	}
	set version1 $r1
	set version2 $r2
    }
    return 0
}

#
# run clock with gmt time zone
#

proc gmtclock {args} {
    # also set TZ environment variable, since -gmt option of "clock"
    #   appears to be off by an hour sometimes; it was observed on a
    #   Solaris 5.3 system with the TZ set to "US/Central" while in
    #   daylight savings time.
    global env
    if {[info exists env(TZ)]} {set svtz $env(TZ)}
    set env(TZ) GMT
    set result [eval clock $args "-gmt 1"]
    if {[info exists svtz]} {set env(TZ) $svtz} else {unset env(TZ)}
    return $result
}

#
# Return the current date and time
#
proc currentTime {} {
    return [httpTime]
}

#
# Return time in preferred http format.  If a parameter is given, it is
#  assumed to be a time in seconds like [clock seconds], otherwise the
#  current time is used.
#
proc httpTime {{seconds ""}} {
    if {$seconds == ""} {
	set seconds [clock seconds]
    }
    return [gmtclock format $seconds -format {%a, %d %b %Y %H:%M:%S GMT}]
}

#
# Return time current time for log file.
#
proc logTime {} {
    return [gmtclock format [clock seconds] -format {%Y/%m/%d %H:%M:%S}]
}

#
# Scan any time format and convert into "seconds" format
#
# Accepts any of the three official HTTP/1.0 formats:
#     Sun, 06 Nov 1994 08:49:37 GMT (preferred)
#     Sunday, 06-Nov-94 08:49:37 GMT
#     Sun Nov  6 08:49:37 1994
# and the old and new style PGP times
#     1994/11/06 08:49 GMT
#     Sun Nov  6 08:49:37 1994 GMT
# and the logTime format (GMT assumed)
#     1994/11/06 08:49:37
# and more, depending on what Tcl's [clock scan] can handle

array set weekdayTable "Sun {} Mon {} Tue {} Wed {} Thu {} Fri {} Sat {}"

proc scanAnyTime {time} {
    global weekdayTable
    set time [string trim $time]
    set tz ""
    if {[info exists weekdayTable([string range $time 0 2])]} {
	# get rid of the day of the week, and turn dashes into spaces
	set time [lrange $time 1 end]
	regsub -all -- "-" $time " " time
    } elseif {[scan $time "%d/%d/%d %d:%d:%d %s" year mon day hour min sec tz] >= 6} {
	set time [format "%02d:%02d:%02d %02d/%02d/%02d %s" \
				$hour $min $sec $mon $day $year $tz]
    } elseif {[scan $time "%d/%d/%d %d:%d %s" year mon day hour min tz] >= 5} {
	set time [format "%02d:%02d %02d/%02d/%02d %s" \
				$hour $min $mon $day $year $tz]
    }
    return [gmtclock scan $time]
}

#
# Canonicalize the time to be the format
#     Sun, 06 Nov 1994 08:49:37 GMT
#
proc canonicalizeTime {time} {
    return [httpTime [scanAnyTime $time]]
}

#
# Compare two times.  Return negative if time1 is less than time2,
#   0 if the times are equal, or positive if time1 is greater than time2.
#
proc compareTimes {time1 time2} {
    set time1seconds [scanAnyTime $time1]
    set time2seconds [scanAnyTime $time2]
    if {$time1seconds < $time2seconds} {
	return -1
    }
    if {$time1seconds > $time2seconds} {
	return 1
    }
    return 0
}
