/*
 * Unix C routines for NSBD to interface with Tcl.
 * File created by Dave Dykstra <dwd@bell-labs.com>, 01 Nov 1996
 */
/*
 * Copyright (C) 1996-7 by Dave Dykstra and Lucent Technologies
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * If those terms are not sufficient for you, contact the author to
 * discuss the possibility of an alternate license.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#include <tcl.h>
#include <time.h>
#include <stdio.h>
#include <signal.h>
#include <malloc.h>

static Tcl_Interp *nsbd_interp;

RETSIGTYPE
#ifdef _USING_PROTOTYPES_
catch_nsbdsig(int sig)
#else
catch_nsbdsig(sig)
    int sig;
#endif
{
    char cmd[80];

    sprintf(cmd, "nsbdExitCleanup; exit %d", sig);
    Tcl_Eval(nsbd_interp, cmd);
}

static void
#ifdef _USING_PROTOTYPES_
setsigs(Tcl_Interp *interp)
#else
setsigs(interp)
    Tcl_Interp *interp;
#endif
{
    nsbd_interp = interp;

    if (signal(SIGHUP, catch_nsbdsig) == SIG_IGN)
	signal(SIGHUP, SIG_IGN);
    if (signal(SIGINT, catch_nsbdsig) == SIG_IGN)
	signal(SIGINT, SIG_IGN);
    if (signal(SIGTERM, catch_nsbdsig) == SIG_IGN)
	signal(SIGTERM, SIG_IGN);
}

static int
#ifdef _USING_PROTOTYPES_
nsbd_startTk(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
#else
nsbd_startTk(clientData, interp, argc, argv)
    ClientData clientData;
    Tcl_Interp *interp;
    int argc;
    char *argv[];
#endif
{
#ifdef STANDALONE
    if (Tk_InitStandAlone(interp) == TCL_ERROR)
	return TCL_ERROR;
    Tcl_StaticPackage(interp, "Tk", Tk_InitStandAlone,
				(Tcl_PackageInitProc *) NULL);
    return TCL_OK;
#else
    static char loadtk[] = {
    /* At least on tcl7.5, the default auto_path when run non-interactively */
    /*   for some strange reason only includes lib/tcl7.5, not lib itself */
    "global auto_path\n\
    if {[lsearch $auto_path [file dirname [info library]]] < 0} {\n\
	lappend auto_path [file dirname [info library]]\n\
     }\n\
     package require Tk"
    };

    if (Tk_Init(interp) == TCL_ERROR)
	return TCL_ERROR;

    return Tcl_Eval(interp, loadtk);
#endif
}

void
#ifdef _USING_PROTOTYPES_
nsbdTclfree(char *blockPtr)
#else
nsbdTclfree(blockPtr)
    char *blockPtr;
#endif
{
    free(blockPtr);
}

static int
#ifdef _USING_PROTOTYPES_
nsbderror(Tcl_Interp *interp)
#else
nsbderror(interp)
    Tcl_Interp *interp;
#endif
{
    Tcl_SetErrorCode(interp, "NSBD", "INTERNAL", (char *) NULL);
    return TCL_ERROR;
}

#if !HAVE_STRERROR
char *
#ifdef _USING_PROTOTYPES_
strerror(int errcode)
#else
strerror(errcode)
	int errcode;
#endif
{
	extern int sys_nerr;
	extern char *sys_errlist[];

	if (errcode < 0 || errcode > sys_nerr)
		return("Unknown error");
	else
		return(sys_errlist[errcode]);
}
#else
extern char * strerror _ANSI_ARGS_((int errcode));
#endif

static int
#ifdef _USING_PROTOTYPES_
nsbd_symLink(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
#else
nsbd_symLink(clientData, interp, argc, argv)
    ClientData clientData;
    Tcl_Interp *interp;
    int argc;
    char *argv[];
#endif
{
    if (argc != 3) {
	interp->result = "wrong # args: should be \"symLink fromFile toFile\"";
	return TCL_ERROR;
    }
    if (symlink(argv[1], argv[2]) == -1) {
	extern int errno;
	char *errmsg = malloc(strlen(argv[2]) + 100);
	sprintf(errmsg, "cannot create %s: %s", argv[2], strerror(errno));
	interp->freeProc = nsbdTclfree;
	interp->result = errmsg;
	return (nsbderror(interp));
    }
    return TCL_OK;
}

/*
 * nsbdnestlevel is here for speed when parsing large files;  it was 
 *   measured to be about 4 times faster than the tcl version on tcl8.0a2
 *   on a file of about 10000 lines; the tcl version took up about half
 *   the parse time of that file, and this C version made it drop to
 *   about 1/5th of the total parse time.
 * nsbdnestlevel figures out the nestlevel in a line of an "nsbd" file.
 * parameters are
 *    1. line - the line to parse
 *    2. linelength - the number of bytes in the line
 *    3. lineno - the line number that is being parsed
 * returns list of two items: the nesting level and the number of characters
 *    of whitespace at the beginning of the line
 */

#define MAXNESTLEVEL 50
static int nestlength[MAXNESTLEVEL+1];
static int nestlevel;

static int
#ifdef _USING_PROTOTYPES_
nsbd_nestlevelInit(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
#else
nsbd_nestlevelInit(clientData, interp, argc, argv)
    ClientData clientData;
    Tcl_Interp *interp;
    int argc;
    char *argv[];
#endif
{
    nestlevel = 0;
    return TCL_OK;
}

static int
#ifdef _USING_PROTOTYPES_
nsbd_nestlevel(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
#else
nsbd_nestlevel(clientData, interp, argc, argv)
    ClientData clientData;
    Tcl_Interp *interp;
    int argc;
    char *argv[];
#endif
{
    char *line, *lineno;
    int linelength;
    int nlength = 0, whitechars = 0;
    int i, c;
    int curlength; 

    if (argc != 4) {
	interp->result = "wrong # args: should be \"nsbdnestlevel line linelength lineno\"";
	return TCL_ERROR;
    }
    
    line = argv[1];
    sscanf(argv[2], "%d", &linelength);
    lineno = argv[3];
    c = 0;
    for (i = 0; i < linelength; i++) {
	if ((c = line[i]) == ' ') {
	    nlength++;
	    whitechars++;
	}
	else if (c == '\t') {
	    nlength += 8 - (nlength % 8);
	    whitechars++;
	}
	else {
	    break;
	}
    }
    if ((c != '#') && (whitechars != linelength)) {
	curlength = nestlength[nestlevel];
	if (nlength != curlength) {
	    if (nlength > curlength) {
		if (++nestlevel >= MAXNESTLEVEL) {
		    sprintf(interp->result,
			    "nesting level too deep at line %s", lineno);
		    return (nsbderror(interp));
		}
		nestlength[nestlevel] = nlength;
	    }
	    else {
		/* must be less, find out which if any previous level matches */
		nestlevel--;
		while ((nestlevel > 0) && (nlength < nestlength[nestlevel])) {
		    nestlevel--;
		}
		if (nlength != nestlength[nestlevel]) {
		    sprintf(interp->result,
			"invalid indent level on line %s", lineno);
		    return(nsbderror(interp));
		}
	    }
	}
    }
    sprintf(interp->result, "%d %d", nestlevel, whitechars);
    return TCL_OK;
}


#ifdef STANDALONE
#define Tcl_Init Tcl_InitStandAlone
#endif

int
#ifdef _USING_PROTOTYPES_
nsbd_init(Tcl_Interp *interp)
#else
nsbd_init(interp)
    Tcl_Interp *interp;
#endif
{
    if (Tcl_Init(interp) == TCL_ERROR)
	return TCL_ERROR;
    setsigs(interp);

#ifdef STANDALONE
    if (Tcl_Eval(interp, "set nsbdStandAlone 1") == TCL_ERROR) {
	return TCL_ERROR;
    }
#endif
    /* Tclgdbm_Init(interp); */
    Tclmd5_Init(interp);
    Tclsha1_Init(interp);

    Tcl_CreateCommand(interp, "startTk", nsbd_startTk,
	(ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateCommand(interp, "symLink", nsbd_symLink,
	(ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateCommand(interp, "nsbdnestlevelInit", nsbd_nestlevelInit,
	(ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateCommand(interp, "nsbdnestlevel", nsbd_nestlevel,
	(ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);

    return TCL_OK;
}

/*
 * being called from tcl2c and a tcl installation with plus patch
 */
int
#ifdef _USING_PROTOTYPES_
nsbd_initStandAlone(Tcl_Interp *interp)
#else
nsbd_initStandAlone(interp)
    Tcl_Interp *interp;
#endif
{
    return(nsbd_init(interp));
}

